from typing import Union, Optional, Type, Tuple
import abc

from fibomat.linalg import VectorLike, Vector, DimVector, VectorValueError, TransformableBase
from fibomat.layout.utils import _check_lattice_vectors


class LatticeBuilderBase(abc.ABC):

    _VectorClass: Type[Union[Vector, DimVector]]

    def __init__(
        self,
        nu: int, nv: int,
        u: Union[VectorLike, DimVector], v: Union[VectorLike, DimVector],
        center: Optional[Union[Vector, DimVector]] = None
    ):
        self._nu = int(nu)
        self._nv = int(nv)

        if self._nu < 1 or self._nv < 1:
            raise ValueError('nu and nv must be at least 1.')

        self._u = self._VectorClass(u)
        self._v = self._VectorClass(v)

        _check_lattice_vectors(self._u, self._v)

        self._offset = -(self._nu-1)/2 * self._u - (self._nv-1)/2 * self._v + self._VectorClass(center)

        self._elements = []
        self._added_sites = set()

    @property
    def nu(self):
        return self.nu

    @property
    def nv(self):
        return self.nv

    def __setitem__(self, key: Tuple[int, int], element: TransformableBase):
        if not isinstance(key, tuple) or not len(key) == 2 or not isinstance(key[0], int) or not isinstance(key[1], int):
            raise TypeError('key must be Tuple[int, int].')

        if key in self._added_sites:
            raise ValueError('Lattice site is already set.')

        if key[0] >= self._nu or key[1] >= self._nv:
            raise ValueError('key[0] >= nu or key[1] >= nv')

        pos = key[0] * self._u + key[1] * self._v + self._offset

        try:
            self._elements.append(element.translated_to(pos))
        except VectorValueError as vec_val_error:
            raise RuntimeError(
                'Maybe you tried to add a dimensioned object to a non dimensioned lattice builder (or vice versa)?'
            ) from vec_val_error

        self._added_sites.add(key)



"""Provides the :class:`RasterizedPoints` class."""
from __future__ import annotations
from typing import List, Optional

import numpy as np

from fibomat.linalg import Vector, VectorLike, BoundingBox
from fibomat.shapes import Shape


class RasterizedPoints(Shape):
    """Class represents pre-rasterized points."""
    def __init__(self, dwell_points: np.ndarray, is_closed: bool, description: Optional[str] = None):
        """
        Args:
            dwell_points (np.ndarray):
                dim must be 2 with shape = (-1, 3). First to components are coordinates and  third a dwell time
                multiplicand.
            is_closed (bool): if True, shape is considered as closed.
            description (str, optional): description

        Raises:
            ValueError: Raised if dimension or shape of dwell points is wrong
        """
        super().__init__(description)

        if dwell_points.ndim != 2:
            raise ValueError('dwell_points mus have dimension 2.')
        if dwell_points.shape[1] != 3:
            raise ValueError('dwell_points must have shape (N, 3)')

        self._dwell_points = dwell_points
        self._is_closed = bool(is_closed)

    @classmethod
    def merged(cls, other_raster_points: List[RasterizedPoints]) -> RasterizedPoints:
        """Create merged :class:`RasterizedPoints` class from list of :class:`RasterizedPoints`.

        Args:
            other_raster_points: points to be merged.

        Returns:
            RasterizedPoints
        """
        total_points = 0
        for raster_points in other_raster_points:
            total_points += raster_points.n_points

        dwell_points = np.empty(shape=(total_points, 3), dtype=float)

        i_offset = 0
        for raster_points in other_raster_points:
            dwell_points[i_offset:i_offset+raster_points.n_points] = raster_points.dwell_points
            i_offset += raster_points.n_points

        return cls(dwell_points, False)

    @property
    def positions(self):
        """Coordinates of dwell points.

        Access:
            get

        Returns:
            np.ndarray
        """
        view = self._dwell_points[:, :2]
        view.flags.writeable = False
        return view

    @property
    def weights(self):
        """Dwell time multiplicator of dwell points.

        Access:
            get

        Returns:
            np.ndarray
        """
        view = self._dwell_points[:, 2]
        view.flags.writeable = False
        return view

    @property
    def dwell_points(self):
        """Dwell points.

        Access:
            get

        Returns:
            np.ndarray
        """
        view = self._dwell_points[:]
        view.flags.writeable = False
        return view

    @property
    def n_points(self) -> int:
        """Number of dwell points.

        Access:
            get

        Returns:
            int
        """
        return len(self._dwell_points)

    @property
    def is_closed(self) -> bool:
        return self._is_closed

    def __repr__(self) -> str:
        return 'RasterizedPoints(...)'

    @property
    def bounding_box(self) -> BoundingBox:
        return BoundingBox.from_points(self.positions)

    @property
    def center(self) -> Vector:
        return Vector(np.mean(self.positions, axis=1))

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._dwell_points[:, :2] += trans_vec

    def _impl_rotate(self, theta: float) -> None:
        # pylint: disable=invalid-name
        theta = float(theta)
        m = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

        self._dwell_points[:, :2] = self._dwell_points[:, 2] @ m.T

    def _impl_scale(self, fac: float) -> None:
        self._dwell_points[:, :2] *= float(fac)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        # pylint: disable=invalid-name
        mirror_axis = Vector(mirror_axis)

        lx, ly = mirror_axis
        mirror_matrix = np.array([[lx*lx - ly*ly, 2*lx*ly], [2*lx*ly, ly*ly - lx*lx]]) / mirror_axis.length

        self._dwell_points[:, :2] = mirror_matrix.dot(self._dwell_points[:, :2].T)

    def repeats_applied(self, repeats: int) -> RasterizedPoints:
        """Return :class:`RasterizedPoints` with dwel points repeated `repeats` times.

        Args:
            repeats: number of repeats.

        Returns:
            RasterizedPoints
        """
        if repeats != 1:
            # TODO: check and replace
            # np.tile(self._dwell_points, (repeats, 1))
            return RasterizedPoints(np.concatenate([self._dwell_points]*repeats), self._is_closed)

        return self



"""Provides the :class:`Line` class."""
from __future__ import annotations
from typing import Optional

from fibomat.linalg import Vector, VectorLike, BoundingBox
from fibomat.shapes import shape

from fibomat.shapes.arc_spline import ArcSpline, ArcSplineCompatible


class Line(shape.Shape, ArcSplineCompatible):
    """1-dim line."""
    def __init__(self, start: VectorLike, end: VectorLike, description: Optional[str] = None):
        """
        Args:
            start (VectorLike): start point of line
            end (VectorLike): end point of line
            description (str, optional): description
        """
        super().__init__(description)

        start = Vector(start)
        end = Vector(end)

        self._line: ArcSpline = ArcSpline([(*start, 0.), (*end, 0.)], is_closed=False)

    def to_arc_spline(self) -> ArcSpline:
        return self._line.clone_with_new_description(self.description)

    @property
    def start(self) -> Vector:
        """Start point.

        Access:
            get

        Returns:
            Vector
        """
        return self._line.start

    @property
    def end(self) -> Vector:
        """End point.

        Access:
            get

        Returns:
            Vector
        """
        return self._line.end

    @property
    def length(self) -> float:
        """length of line.

        Access:
            get

        Returns:
            float
        """
        return self._line.length

    # def clone(self) -> Line:
    #     return self.__class__(self._line.clone())
    #
    # __copy__ = clone

    def __repr__(self) -> str:
        return '{}(start={!r},end={!r}'.format(
            self.__class__.__name__, self.start, self.end)

    @property
    def is_closed(self) -> bool:
        return False

    @property
    def bounding_box(self) -> BoundingBox:
        return self._line.bounding_box

    @property
    def center(self) -> Vector:
        return self._line.center

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._line._impl_translate(trans_vec)  # pylint: disable=protected-access

    def _impl_rotate(self, theta: float) -> None:
        self._line._impl_rotate(theta)  # pylint: disable=protected-access

    def _impl_scale(self, fac: float) -> None:
        self._line._impl_scale(fac)  # pylint: disable=protected-access

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._line._impl_mirror(mirror_axis)  # pylint: disable=protected-access



from __future__ import annotations
from typing import Tuple

import numpy as np


def _make_fixed_tangent_curvature(
    parametric_curve, param: float, epsilon: float, side: str
) -> Tuple[np.ndarray, float]:
    t = parametric_curve.df(param)
    k = parametric_curve.curvature(param)

    if np.allclose(t, 0):
        if side == 'left':
            t = parametric_curve.df(param + epsilon)
            k = parametric_curve.curvature(param + epsilon)
        elif side == 'right':
            t = parametric_curve.df(param - epsilon)
            k = parametric_curve.curvature(param - epsilon)
        else:
            raise ValueError(f'unknown side {side}')

        assert not np.allclose(t, 0)

    return t, k


def _make_perp_vector(p):
    return np.array([-p[1], p[0]], dtype=p.dtype if isinstance(p, np.ndarray) else float)


class _LinearFunction:
    def __init__(self, p, p_sup):
        self.p = np.array(p)
        self.p_sup = np.array(p_sup)

    def __repr__(self):
        return 't |-> {} + t * {}'.format(self.p_sup, self.p)

    def __call__(self, u):
        return self.p_sup + u*self.p

    def parallel_to(self, other: _LinearFunction):
        return np.isclose(
            np.dot(
                self.p / np.linalg.norm(self.p),
                other.p / np.linalg.norm(other.p)
            ), 1.
        )

    def intersect_at(self, other: _LinearFunction):
        m = np.array([self.p, -other.p]).T
        b = other.p_sup - self.p_sup

        res = np.linalg.solve(m, b)

        return self.p_sup + res[0] * self.p


class _PerpBisector(_LinearFunction):
    def __init__(self, p_0, p_1):
        super().__init__(_make_perp_vector(p_1 - p_0), (p_0 + p_1) / 2)


class _Bisector(_LinearFunction):
    def __init__(self, p_0, p_1):
        p_0 = np.array(p_0)
        p_1 = np.array(p_1)
        super().__init__(p_1 - p_0, (p_0 + p_1) / 2)


def _circle_circle_intersection(circ_1, circ_2):
    c_0, r_0 = circ_1
    c_1, r_1 = circ_2

    c_0 = np.array(c_0)
    c_1 = np.array(c_1)

    d = np.linalg.norm(c_0 - c_1)

    if d > r_0 + r_1:
        raise RuntimeError
    elif d < abs(r_0 - r_1):
        raise RuntimeError
    elif np.isclose(d, 0.) and np.isclose(r_0, r_1):
        return tuple()
    else:
        a = (r_0**2 - r_1**2 + d**2) / (2 * d)

        h = np.sqrt(r_0**2 - a**2)

        c = c_0 + a * (c_1 - c_0) / d

        p_1 = np.array([
            c[0] + h * (c_1[1] - c_0[1]) / d,
            c[1] - h * (c_1[0] - c_0[0]) / d
        ])

        p_2 = np.array([
            c[0] - h * (c_1[1] - c_0[1]) / d,
            c[1] + h * (c_1[0] - c_0[0]) / d
        ])

        if np.allclose(p_1, p_2):
            return p_1,
        else:
            return p_1, p_2



"""Provides the :class:`ArcSpline` and :class:`ArcSplineCompatible` classes."""
from __future__ import annotations
from typing import Optional, Union, Sequence, Iterable, Protocol, runtime_checkable, Tuple, List, Dict

import numpy as np

from fibomat.linalg import VectorLike, Vector, BoundingBox
from fibomat.shapes.shape import Shape

from fibomat import _libfibomat


@runtime_checkable
class ArcSplineCompatible(Protocol):  # pylint: disable=too-few-public-methods
    """Abstract class can be used to mark ArcSpline compatible shapes."""
    def to_arc_spline(self) -> ArcSpline:
        """
        Transform shape to ArcSpline.

        Returns:
            ArcSpline
        """
        raise NotImplementedError


class ArcSpline(Shape, ArcSplineCompatible):
    """Class represents a spline containing circular arcs and straight line segments. The spline is C^0, hence,
    continuous but not differentiable.
    """
    def __init__(
            self,
            arc_spline: Union[_libfibomat.ArcSpline, np.ndarray],
            is_closed: Optional[bool] = None,
            description: Optional[str] = None):
        """
        Args:
            arc_spline (_libfibomat.ArcSpline, np.ndarray):
                np.ndarray must have shape = (-1, 3) where each point is given by (x, y, bulge).
            is_closed (bool, optional):
                if True, the last and first point are connected (potentially with an arc, if the bulge value of the last
                vertex is nonzero). if False, the bulge value of the last point is ignored. The argumetn must be given
                if arc_spline is np.ndarray and is ignored if arc_spline is _libfibomat.ArcSpline.
            description (str, optional): description

        Raises:
            ValueError: Raised if arc_spline is np.ndarray but is_closed not given.
        """
        super().__init__(description)

        if isinstance(arc_spline, _libfibomat.ArcSpline):
            self._arc_spline = _libfibomat.ArcSpline(arc_spline)
        else:
            if is_closed is None:
                raise ValueError('is_closed must be defined if ArcSpline is build from vertices.')
            self._arc_spline = _libfibomat.ArcSpline(arc_spline, is_closed)

        # self._vertices = np.array(self._arc_spline.vertices)
        # self._vertices.flags.writeable = False

    def __copy__(self):
        return self.__class__(arc_spline=self._arc_spline.clone(), description=self.description)

    def __deepcopy__(self, memodict):
        return self.__copy__()

    def __len__(self) -> int:
        return self._arc_spline.size

    # shape.Shape methods

    @classmethod
    def from_segments(cls, segments: Iterable[ArcSplineCompatible], description: Optional[str] = None):
        """Build an ArcSpline from connected segments.

        Args:
            segments (Iterable[ArcSplineCompatible]):
                segments. the start and end point of to consecutive segments must be equal. If this also holds for the
                last and first segment, the curve ist closed.
            description (str, optional): description.

        Returns:
            ArcSpline

        Raises:
            RuntimeError: Raised if segments are not connected.
        """
        vertices: List[np.ndarray] = []

        for i_seg, seg in enumerate(segments):

            arc_spline = seg.to_arc_spline()

            if arc_spline.is_closed and i_seg > 0:
                raise RuntimeError('Cannot build ArcSpline from segments because some segments are closed.')

            seg_vertices = np.array(arc_spline.vertices)

            if vertices:
                if not np.allclose(vertices[-1][-1, :2], seg_vertices[0, :2]):
                    raise RuntimeError('Segments are not C^0. The distance is {}'.format(
                        np.linalg.norm(vertices[-1][-1, :2] - seg_vertices[0, :2]))
                    )
                vertices[-1][-1] = seg_vertices[0]
                vertices.append(seg_vertices[1:])
            else:
                vertices.append(seg_vertices)

        conc_vertices: np.ndarray = np.concatenate(vertices)

        if np.allclose(conc_vertices[0, :2], conc_vertices[-1, :2]):
            return cls(conc_vertices[:-1], True, description)

        return cls(conc_vertices, False, description)

    @classmethod
    def from_shape(cls, segment: ArcSplineCompatible):
        """Converts a single segment to an ArcSpline.

        Args:
            segment (ArcSplineCompatible): segment.

        Returns:
            ArcSpline
        """
        return segment.to_arc_spline()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(start={self.start}, end={self.end}, description={self.description})'

    @property
    def is_closed(self) -> bool:
        return self._arc_spline.is_closed

    def to_arc_spline(self) -> ArcSpline:
        return self

    # Transformable methods

    def clone(self) -> ArcSpline:
        return self.__class__(arc_spline=self._arc_spline.clone(), description=self.description, )

    def clone_with_new_description(self, description: Optional[str] = None):
        """Similar to :meth:`ArcSpline.clone` but set the description the passed description.

        Args:
            description (str, optional): description.

        Returns:
            ArcSpline
        """
        return self.__class__(arc_spline=self._arc_spline.clone(), description=description)

    @property
    def bounding_box(self) -> BoundingBox:
        # return BoundingBox(*self._arc_spline.bounding_box)

        # there is a bug in cavc bounding box calculation for multiple arc segments (cf. examples/fill_lines.py without this hacky code)

        segments = self.segments

        if len(segments) > 1:
            bbox = segments[0].bounding_box
            for seg in segments[1:]:
                bbox = bbox.extended(seg.bounding_box)
            
            return bbox
        else:
            return BoundingBox(*self._arc_spline.bounding_box)

    @property
    def center(self) -> Vector:
        return Vector(self._arc_spline.center)

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._arc_spline.impl_translate(trans_vec)

    def _impl_scale(self, fac: float) -> None:
        self._arc_spline.impl_scale(fac)

    def _impl_rotate(self, theta: float) -> None:
        self._arc_spline.impl_rotate(theta)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._arc_spline.impl_mirror(mirror_axis)

    # other utility methods

    @property
    def start(self) -> Vector:
        """Start point curve

        Access:
            get

        Returns:
            Vector
        """
        return Vector(self._arc_spline.start[:2])

    @property
    def end(self) -> Vector:
        """End point curve

        Access:
            get

        Returns:
            Vector
        """
        if self._arc_spline.is_closed:
            return Vector(self._arc_spline.start[:2])

        return Vector(self._arc_spline.end[:2])

    @property
    def vertices(self) -> np.ndarray:
        """Curve vertices.

        Access:
            get

        Returns:
            np.ndarray
        """
        return np.array(self._arc_spline.vertices)

    @property
    def segments(self) -> Sequence[ArcSplineCompatible]:
        """Return a list of Line and Arc elements representing the curve.

        .. note:: This method is not bijective with regard to the added shapes. E.g. if the curve was constructed from a
                  Circle, this method will not return a Circle element but two Arcs.

        Access:
            get

        Returns:
            List[shape.Shape]
        """
        from fibomat.shapes.line import Line  # pylint: disable=import-outside-toplevel
        from fibomat.shapes.arc import Arc  # pylint: disable=import-outside-toplevel

        def _make_segment(start_vertex, end_vertex):
            if np.isclose(start_vertex[2], 0.):
                return Line(start_vertex[:2], end_vertex[:2])

            return Arc.from_bulge(start_vertex[:2], end_vertex[:2], start_vertex[2])

        segments = []
        vertices = self.vertices
        for i, vertex in enumerate(vertices[1:], start=1):
            segments.append(_make_segment(vertices[i-1], vertex))

        if self.is_closed:
            segments.append(_make_segment(vertices[-1], vertices[0]))

        return segments

    @property
    def arc_spline_impl(self) -> _libfibomat.ArcSpline:
        return self._arc_spline

    @property
    def orientation(self) -> bool:
        """Orientation of curve. True if curve is counterclockwise.

        .. note:: This property is only defined for closed curves.

        Access:
            get

        Returns:
            bool
        """
        return self._arc_spline.orientation

    @property
    def length(self) -> float:
        """Length of curve.

        Access:
            get

        Returns:
            float
        """
        return self._arc_spline.length

    def contains(self, pos: VectorLike):
        pos = Vector(pos)
        return self._arc_spline.contains(pos.x, pos.y)

    def closest_point(self, pos: VectorLike) -> Dict:
        pos = Vector(pos)
        res = self._arc_spline.closest_point(pos.x, pos.y)

        return {'segment': pos[0], 'point': Vector(res[1]), 'distance': res[2]}

    def unit_tangents(self, i_vertex: int) -> Tuple[Optional[Vector], Optional[Vector]]:
        """Unit tangents at vertex i_vertex.

        Args:
            i_vertex (int): vertex index

        Returns:
            Tuple[Optional[Vector], Optional[Vector]]:
                left tangent, right tangent. If any of the two tangents does not exist, the tuple entry will be None.

        Raises:
            ValueError: Raised if i_vertex > #segmens.
        """
        from fibomat.shapes.arc import Arc  # pylint: disable=import-outside-toplevel

        if i_vertex < self._arc_spline.size:
            vertices = self.vertices
            vertex = vertices[i_vertex]

            first_tangent = None
            second_tangent = None

            if i_vertex != 0 or self.is_closed:
                vertex_before = vertices[(i_vertex-1) % len(vertices)]

                if np.isclose(vertex_before[2], 0.):
                    first_tangent = vertex[:2] - vertex_before[:2]
                    first_tangent /= np.linalg.norm(first_tangent)
                else:
                    arc = Arc.from_bulge(vertex_before[:2], vertex[:2], vertex_before[2])
                    first_tangent = arc.unit_tangent_end

            if i_vertex != self._arc_spline.size - 1 or self.is_closed:
                vertex_next = vertices[(i_vertex+1) % len(vertices)]

                if np.isclose(vertex[2], 0.):
                    second_tangent = vertex_next[:2] - vertex[:2]
                    second_tangent /= np.linalg.norm(second_tangent)
                else:
                    arc = Arc.from_bulge(vertex[:2], vertex_next[:2], vertex[2])
                    second_tangent = arc.unit_tangent_start

            return first_tangent, second_tangent

        raise ValueError('i_vertex >= number of segments.')

    def kinks(self) -> List[int]:
        """Return kinks (non differentiable points) of the spline.

        Returns:
             List[int]: vertex indices of kinks.
        """
        tangents = [self.unit_tangents(i) for i in range(len(self.vertices))]

        kinks = []

        if self.is_closed:
            if not np.allclose(tangents[0][0], tangents[0][1]):
                kinks.append(0)

        for i in range(1, len(self.vertices) - 1):
            if not np.allclose(tangents[i][0], tangents[i][1]):
                kinks.append(i)

        if self.is_closed:
            if not np.allclose(tangents[-1][0], tangents[-1][1]):
                kinks.append(len(self.vertices) - 1)

        return kinks

    def segments_at_vertex(self, i_vertex: int) -> Tuple[Optional[Shape], Optional[Shape]]:
        """Return the segments around the vertex with index i_vertex.

        Args:
            i_vertex: vertex index

        Returns:
            Tuple[Optional[Shape], Optional[Shape]]:
                left and right segments. If any of the segments is not defined, the tuple entry will be None.

        Raises:
            ValueError: Raised if i_vertex > #segmens.
        """
        from fibomat.shapes.arc import Arc  # pylint: disable=import-outside-toplevel
        from fibomat.shapes.line import Line  # pylint: disable=import-outside-toplevel

        if i_vertex < self._arc_spline.size:
            vertices = self.vertices
            vertex = vertices[i_vertex]

            first_seg = None
            second_seg = None

            if i_vertex != 0 or self.is_closed:
                vertex_before = vertices[(i_vertex-1) % len(vertices)]

                if np.isclose(vertex_before[2], 0.):
                    first_seg = Line(vertex_before[:2], vertex[:2])
                else:
                    first_seg = Arc.from_bulge(vertex_before[:2], vertex[:2], vertex_before[2])

            if i_vertex != self._arc_spline.size - 1 or self.is_closed:
                vertex_next = vertices[(i_vertex+1) % len(vertices)]

                if np.isclose(vertex[2], 0.):
                    second_seg = Line(vertex[:2], vertex_next[:2])
                else:
                    second_seg = Arc.from_bulge(vertex[:2], vertex_next[:2], vertex[2])

            return first_seg, second_seg

        raise ValueError('i_vertex >= number of segments.')

    def reversed(self):
        """Return a reversed copy of the arc spline

        Returns:
            ArcSpline
        """
        clone = self.clone()
        clone._arc_spline.reverse()  # pylint: disable=protected-access
        return clone



from __future__ import annotations

import numpy as np

from fibomat.linalg.vectors import VectorLike, Vector


def make_perp_vector(vec: VectorLike) -> Vector:
    vec = Vector(vec)
    return Vector(-vec.y, vec.x)


class GeomLine:
    def __init__(self, direction: VectorLike, support: VectorLike):
        self._direction = np.asarray(Vector(direction))
        self._support = np.asarray(Vector(support))

        if np.allclose(self._direction, 0.):
            raise RuntimeError('direction must not be the null vector.')

    @classmethod
    def make_bisector(cls, p_0: VectorLike, p_1: VectorLike):
        p_0 = Vector(p_0)
        p_1 = Vector(p_1)

        return cls(p_1 - p_0, (p_0 + p_1) / 2)

    @classmethod
    def make_perp_bisector(cls, p_0: VectorLike, p_1: VectorLike):
        p_0 = Vector(p_0)
        p_1 = Vector(p_1)

        return cls(make_perp_vector(p_1 - p_0), (p_0 + p_1) / 2)

    def __call__(self, u):
        return self._support + u * self._direction

    def parallel_to(self, other: GeomLine):
        return np.isclose(
            np.dot(
                self._direction / np.linalg.norm(self._direction),
                other._direction / np.linalg.norm(other._direction)
            ), 1.
        )

    def intersect_at_param(self, other: GeomLine):
        m = np.array([self._direction, -other._direction]).T
        b = other._support - self._support

        res = np.linalg.solve(m, b)

        return res[0]

    def intersect_at(self, other: GeomLine):
        res = self.intersect_at_param(other)

        return self._support + res * self._direction

    def find_param(self, point: VectorLike):
        point = Vector(point)

        if np.isclose(self._direction[0], 0.):
            if not np.isclose(point.x, self._support[0]):
                raise ValueError('point lies not on GeomLine.')
            s = (point.y - self._support[1]) / self._direction[1]
        elif np.isclose(self._direction[1], 0.):
            if not np.isclose(point.y, self._support[1]):
                raise ValueError('point lies not on GeomLine.')
            s = (point.x - self._support[0]) / self._direction[0]
        else:
            s_0 = (point.x - self._support[0]) / self._direction[0]
            s_1 = (point.y - self._support[1]) / self._direction[1]

            if not np.isclose(s_0, s_1):
                raise ValueError('point lies not on GeomLine.')

            s = s_0

        return s



from PyQt5 import QtWidgets, QtCore


# https://stackoverflow.com/a/52690011, CC BY-SA 4.0
class Slider(QtWidgets.QSlider):
    def mousePressEvent(self, event):
        super(Slider, self).mousePressEvent(event)
        if event.button() == QtCore.Qt.LeftButton:
            val = self.pixelPosToRangeValue(event.pos())
            self.setValue(val)

    def pixelPosToRangeValue(self, pos):
        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        gr = self.style().subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderGroove, self)
        sr = self.style().subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderHandle, self)

        if self.orientation() == QtCore.Qt.Horizontal:
            sliderLength = sr.width()
            sliderMin = gr.x()
            sliderMax = gr.right() - sliderLength + 1
        else:
            sliderLength = sr.height()
            sliderMin = gr.y()
            sliderMax = gr.bottom() - sliderLength + 1;
        pr = pos - sr.center() + sr.topLeft()
        p = pr.x() if self.orientation() == QtCore.Qt.Horizontal else pr.y()
        return QtWidgets.QStyle.sliderValueFromPosition(self.minimum(), self.maximum(), p - sliderMin,
                                                        sliderMax - sliderMin, opt.upsideDown)



"""Provides the :class:`Spot` class."""
from __future__ import annotations
from typing import Optional

from fibomat.linalg import VectorLike, Vector, BoundingBox
from fibomat.shapes import shape


class Spot(shape.Shape):
    """0-dim spot."""

    def __init__(self, position: Optional[VectorLike] = None, *, description: Optional[str] = None):
        """
        Args:
            position (VectorLike, optional): position of the spot, default to (0, 0)
            description (str, optional): description
        """
        super().__init__(description)

        self._position: Vector = Vector(position) if position is not None else Vector()

    def __repr__(self) -> str:
        return '{}(position={!r})'.format(
            self.__class__.__name__, self._position)

    @property
    def position(self) -> Vector:
        """Position of the spot. Same as :attr:`center`

        Access:
            get

        Returns:
            Vector
        """
        return self._position

    @property
    def bounding_box(self) -> BoundingBox:
        return BoundingBox(self._position, self._position)

    @property
    def is_closed(self) -> bool:
        raise NotImplementedError

    @property
    def center(self) -> Vector:
        return self._position

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._position += trans_vec

    def _impl_rotate(self, theta: float) -> None:
        self._position = self._position.rotated(theta)

    def _impl_scale(self, fac: float) -> None:
        self._position *= fac

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._position = self._position.mirrored(mirror_axis)



from __future__ import annotations
from typing import Generic, TypeVar, Iterable, Union, Type
import abc
import copy

import numpy as np

from fibomat.linalg.vectors import Vector, DimVector, FloatTypes, VectorValueError
from fibomat.units import LengthQuantity


VectorT = TypeVar('VectorT', Vector, DimVector)
ScalarT = TypeVar('ScalarT', int, float, np.integer, np.float, LengthQuantity)
SelfT = TypeVar('SelfT', bound='BoundingBoxBase')


class BoundingBoxBase(Generic[VectorT, ScalarT], abc.ABC):

    _VectorClass: Type[VectorT]

    def __init__(self, lower_left: np.ndarray, upper_right: np.ndarray):
        self._lower_left = self._VectorClass(lower_left)
        self._upper_right = self._VectorClass(upper_right)

        self._is_valid()

    @classmethod
    @abc.abstractmethod
    def from_points(cls, points: Iterable[VectorT]) -> BoundingBoxBase[VectorT]:
        """
        Constructs rectangular bounding box containing all `points`

        Args:
            points (VectorArrayLike): points which should be included in bounding box
            description (str, optional): description

        Returns:
            (BoundingBox): new `BoundingBox`
        """
        raise NotImplementedError

    def _is_valid(self):
        if self._lower_left.x > self._upper_right.x \
           or self._lower_left.y > self._upper_right.y:
            raise ValueError('Invalid coordinates for bounding box')

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BoundingBoxBase):
            raise NotImplementedError
        return self._lower_left.close_to(other._lower_left) and self._upper_right.close_to(other._upper_right)

    def close_to(self, other: object) -> bool:
        return self == other

    def __repr__(self):
        return 'BoundingBox(lower_left=({}, {}), upper_right=({}, {}))'.format(
            self._lower_left.x, self._lower_left.y, self._upper_right.x, self._upper_right.y
        )

    @property
    def lower_left(self) -> VectorT:
        """
        Vector: lower left coordinate

        Access:
            get
        """
        return self._lower_left

    @property
    def upper_right(self) -> VectorT:
        """
        Vector: upper richt coordinate

        Access:
            get
        """
        return self._upper_right

    @property
    def width(self) -> ScalarT:
        """
        float: with of bounding box

        Access:
            get
        """
        return self._upper_right[0] - self._lower_left[0]

    @property
    def height(self) -> ScalarT:
        """
        float: with of bounding box

        Access:
            get
        """
        return self._upper_right[1] - self._lower_left[1]

    @property
    def area(self) -> ScalarT:
        """
        float: area of bounding box

        Access:
            get
        """
        return self.width * self.height

    @property
    def center(self) -> VectorT:
        """
        Vector: center bounding box

        Access:
            get
        """
        return (self._upper_right + self._lower_left) / 2

    def clone(self: SelfT) -> SelfT:
        """
        Clones the bounding box

        Returns:
            (BoundingBox): cloned box
        """
        return copy.deepcopy(self)

    def scaled(self: SelfT, val: float) -> SelfT:
        """
        Return a scaled version of the `BoundingBox`.

        Args:
            val: scale factor

        Returns:
            BoundingBox
        """
        val = float(val)/4

        shift = (self.width * val, self.height * val)

        return self.__class__(self._lower_left - shift, self._upper_right + shift)

    def overlaps_with(self, other: BoundingBoxBase) -> bool:
        """
        Checks if this bounding box overlaps with other.

        Args:
            other: other bounding box

        Returns:
            bool: True if self and other overlap
        """

        overlap_x = True
        overlap_y = True

        if self._lower_left.x > other._upper_right.x or self._upper_right.x < other._lower_left.x:
            overlap_x = False

        if self._lower_left.y > other._upper_right.y or self._upper_right.y < other._lower_left.y:
            overlap_y = False

        return overlap_x and overlap_y

    def contains(self, other: Union[BoundingBoxBase, VectorT]) -> bool:
        """
        Checks if this bounding box contains other. Alternatively, you can use `in` syntax. Note, that a box is contains
        always itself. ::

            box_1 = BoundingBox([1, 2], [3, 4])
            box_2 = BoundingBox([1, 3], [0, 4])

            print(box_1.contains(box_2)
            # or
            print(box_2 in box_1)

        Args:
            other (BoundingBox): other box

        Returns:
            (bool): True if `self` contains `other`
        """
        if isinstance(other, self.__class__):
            if self._lower_left.x <= other._lower_left.x \
               and self._lower_left.y <= other._lower_left.y \
               and self._upper_right.x >= other._upper_right.x \
               and self._upper_right.y >= other._upper_right.y:
                return True
            return False
        else:
            try:
                other = self._VectorClass(other)

                return (
                    self._lower_left.x <= other.x <= self._upper_right.x
                    and self._lower_left.y <= other.y <= self._upper_right.y
                )
            except VectorValueError as error:
                raise TypeError('other must be a instance of BoundingBox or VectorT.') from error

    def __contains__(self, other: BoundingBoxBase) -> bool:
        return self.contains(other)

    # TODO: allow list of vectors here.
    def extended(self: SelfT, other: Union[BoundingBoxBase, VectorT, Iterable[VectorT]]) -> SelfT:
        """
        Return a extended bounding box so that `self` and other are contained

        Args:
            other (BoundingBox): other box
        """

        def _extend_with_vector(bbox, vector) -> None:
            if vector.x < bbox._lower_left.x:
                bbox._lower_left = self._VectorClass(vector.x, bbox._lower_left.y)
            if vector.x > bbox._upper_right.x:
                bbox._upper_right = self._VectorClass(vector.x, bbox._upper_right.y)
            if vector.y < bbox._lower_left.y:
                bbox._lower_left = self._VectorClass(bbox._lower_left.x, vector.y)
            if vector.y > bbox._upper_right.y:
                bbox._upper_right = self._VectorClass(bbox._upper_right.x, vector.y)

        clone = self.clone()

        if isinstance(other, self.__class__):
            if other not in clone:
                if other._lower_left.x < clone._lower_left.x:
                    clone._lower_left = self._VectorClass(other._lower_left.x, clone._lower_left.y)
                if other._lower_left.y < clone._lower_left.y:
                    clone._lower_left = self._VectorClass(clone._lower_left.x, other._lower_left.y)
                if other._upper_right.x > clone._upper_right.x:
                    clone._upper_right = self._VectorClass(other._upper_right.x, clone._upper_right.y)
                if other._upper_right.y > clone._upper_right.y:
                    clone._upper_right = self._VectorClass(clone._upper_right.x, other._upper_right.y)
        else:
            try:
                other = self._VectorClass(other)
            except VectorValueError:
                # other is maybe an Iterable[VectorT]
                try:
                    for p in other:
                        _extend_with_vector(clone, self._VectorClass(p))
                except (TypeError, VectorValueError) as error:
                    raise TypeError('other must be SelfT, VectorT or Iterable[VectorT].') from error
            else:
                _extend_with_vector(clone, other)

        return clone




from fibomat.shapes.polyline import Polyline
from typing import Optional, Dict
import io

import PIL.Image

import numpy as np

import svgwrite
import svgwrite.container
import svgwrite.shapes
import svgwrite.extensions

from fibomat.backend import BackendBase
from fibomat.site import Site
from fibomat.units import U_, Q_, scale_factor
from fibomat.utils import PathLike
from fibomat.pattern import Pattern
from fibomat.shapes import Rect, Polygon, Circle
from fibomat.linalg import DimBoundingBox, scale, translate, Vector


class SVGBackend(BackendBase):
    def __init__(self, pixel_scale: Q_ = Q_('0.01 µm'), stroke_width: float = 0.1, description: Optional[str] = None):
        super().__init__()

        self._description = description
        self._pixel_scale = pixel_scale
        self._stroke_width = stroke_width

        # sites are mapped to svg groups
        self._layers = []
        self._current_layer: Optional[Dict] = None
        self._current_layer_center: Optional[Vector] = None

        self._total_bounding_box = None

    def process_site(self, new_site: Site) -> None:
        self._current_layer = {'id': new_site.description or None, 'elements': []}
        self._current_layer_center = new_site.center.vector_as(self._pixel_scale.u)
        self._layers.append(self._current_layer)

        if not self._total_bounding_box:
            self._total_bounding_box = new_site.fov_bounding_box
        else:
            self._total_bounding_box = self._total_bounding_box.extended(new_site.fov_bounding_box)

        super().process_site(new_site)

    # def process_pattern(self, ptn: Pattern) -> None:
    #     super().process_pattern(ptn)

    def _fill_or_stroke(self, svg_elem, raster_style):
        if raster_style.dimension == 2:
            svg_elem.fill('black')
        elif raster_style.dimension == 1:
            svg_elem.stroke('black', width=self._stroke_width)
            svg_elem['style'] = 'fill:none'
        else:
            raise RuntimeError('Only raster styles with dimension 1 or 2 are supported.')

    def _pixel_scale_factor(self, other_unit):
        return scale_factor(self._pixel_scale, other_unit) / self._pixel_scale.m

    def _scale_and_shift_shape(self, ptn: Pattern):
        fak = self._pixel_scale_factor(ptn.dim_shape.unit)

        return ptn.dim_shape.shape.transformed(scale(fak) | translate(self._current_layer_center))

    def rect(self, ptn: Pattern[Rect]) -> None:
        scale = self._pixel_scale_factor(ptn.dim_shape.unit)

        scaled_rect: Rect = self._scale_and_shift_shape(ptn)

        svg_rect = svgwrite.shapes.Rect(
            insert=scaled_rect.center + (-scaled_rect.width/2, -scaled_rect.height/2),
            size=(scaled_rect.width, scaled_rect.height)
        )

        svg_rect.rotate(np.rad2deg(scaled_rect.theta), center=scaled_rect.center)

        self._fill_or_stroke(svg_rect, ptn.raster_style)

        self._current_layer['elements'].append(svg_rect)

    def polygon(self, ptn: Pattern[Polygon]) -> None:
        scaled_polygon: Polygon = self._scale_and_shift_shape(ptn)

        svg_poly = svgwrite.shapes.Polygon(scaled_polygon.points)

        self._fill_or_stroke(svg_poly, ptn.raster_style)

        self._current_layer['elements'].append(svg_poly)

    def polyline(self, ptn: Pattern[Polyline]) -> None:
        scaled_polyline: Polyline = self._scale_and_shift_shape(ptn)

        svg_poly = svgwrite.shapes.Polyline(scaled_polyline.points)

        self._fill_or_stroke(svg_poly, ptn.raster_style)

        self._current_layer['elements'].append(svg_poly)

    def circle(self, ptn: Pattern[Circle]) -> None:
        scaled_circle: Circle = self._scale_and_shift_shape(ptn)

        svg_circle = svgwrite.shapes.Circle(center=scaled_circle.center, r=scaled_circle.r)

        self._fill_or_stroke(svg_circle, ptn.raster_style)

        self._current_layer['elements'].append(svg_circle)

    def _to_svg(self) -> svgwrite.Drawing:
        if not self._total_bounding_box:
            raise RuntimeError('Site may not be empty.')

        total_width = self._total_bounding_box.width
        total_height = self._total_bounding_box.height

        total_width_scaled = self._pixel_scale_factor(total_width.u) * total_width.m
        total_height_scaled = self._pixel_scale_factor(total_height.u) * total_height.m

        center = self._total_bounding_box.center

        shift_x = self._pixel_scale_factor(center.x.u) * center.x.m
        shift_y = self._pixel_scale_factor(center.y.u) * center.y.m

        svg = svgwrite.Drawing(size=(total_width_scaled, total_height_scaled))
        inkscape_svg = svgwrite.extensions.Inkscape(svg)

        for layer in self._layers:
            svg_layer = inkscape_svg.layer(label=layer['id'] or 'Layer')
            svg.add(svg_layer)
            svg_layer.translate(total_width_scaled/2 - shift_x, total_height_scaled/2 - shift_y)
            svg_layer.scale(sx=1, sy=-1)

            for elem in layer['elements']:
                svg_layer.add(elem)

        return svg

    def save(self, filename: PathLike) -> None:
        self._to_svg().saveas(filename, pretty=True)

    def to_pillow_image(self, background_color=(255, 255, 255, 255)):
        # https://github.com/manatools/dnfdragora/blob/acaa41e511c3ce026a9123fe494bd017cbfb99db/dnfdragora/updater.py

        try:
            import cairosvg
        except Exception as e:
            raise RuntimeError('Please install cairosvg manually') from e

        rendered_svg = PIL.Image.open(io.BytesIO(cairosvg.svg2png(bytestring=self._to_svg().tostring())))

        image = PIL.Image.new('RGBA', size=rendered_svg.size, color=background_color)
        image.paste(rendered_svg, None, rendered_svg)
        image.convert('RGB')

        return image



"""
Provides the :class:`BackendBaseMeta` class.
"""
# pylint: disable=bad-mcs-classmethod-argument


class BackendBaseMeta(type):
    """Metaclass for :class:`BackendBase` to automatically detected available
    and implemented shape methods.
    """
    def __new__(mcs, name, bases, attrs):  # **kwargs
        def get_abstract_shape_methods(method_names, attrs_):
            abstract_shape_methods_ = {}

            for method_name in method_names:
                if not method_name.startswith('__'):
                    is_abstract_shape_method = getattr(
                        attrs[method_name], '_abstract_shape_method', False
                    )
                    if is_abstract_shape_method:
                        shape_type = getattr(
                            attrs[method_name], '_shape_type', False
                        )

                        # TODO: do not use __name__ here
                        abstract_shape_methods_[shape_type] = attrs_[method_name].__name__

            return abstract_shape_methods_

        def get_implemented_shape_methods(method_names, attrs_, abstract_shape_methods_):
            # would be nicer: https://pypi.org/project/bidict/
            abstract_shape_methods_rev = dict(map(reversed, abstract_shape_methods_.items()))
            implemented_shapes = {}
            for method_name in method_names:
                if method_name in abstract_shape_methods_.values():
                    is_abstract_shape_method = getattr(
                        attrs_[method_name], '_abstract_shape_method', False
                    )

                    if not is_abstract_shape_method:
                        implemented_shapes[
                            abstract_shape_methods_rev[method_name]
                        ] = attrs_[method_name]

            return implemented_shapes

        # available shapes
        abstract_shape_methods = {}

        for base in bases:
            abstract_shape_methods.update(
                base._abstract_shape_methods
                # get_abstract_shape_methods(
                #     [attr for attr, _ in inspect.getmembers(base)],
                #     base.__dict__
                # )
            )

        abstract_shape_methods.update(
            get_abstract_shape_methods(attrs.keys(), attrs)
        )

        attrs['_abstract_shape_methods'] = abstract_shape_methods

        # implemented shapes
        implemented_shape_methods = {}

        for base in bases:
            implemented_shape_methods.update(
                base._implemented_shape_methods
                # get_implemented_shape_methods(
                #     [attr for attr, _ in inspect.getmembers(base)],
                #     base.__dict__,
                #     abstract_shape_methods
                # )
            )

        implemented_shape_methods.update(
            get_implemented_shape_methods(attrs.keys(), attrs, abstract_shape_methods)
        )

        attrs['_implemented_shape_methods'] = implemented_shape_methods

        return super().__new__(mcs, name, bases, attrs)



from typing import Optional

from fibomat.layout.lattices.lattice_builder_base import LatticeBuilderBase
from fibomat.layout.lattices.dim_lattice import DimLattice
from fibomat.linalg import VectorLike, DimVector, DimVectorLike
from fibomat.units import LengthQuantity, U_, Q_, has_length_dim


class DimLatticeBuilder(LatticeBuilderBase):
    _VectorClass = DimVector

    def __init__(self, nu: int, nv: int, u: DimVectorLike, v: DimVectorLike, center: Optional[DimVectorLike] = None):
        super().__init__(nu, nv, u, v, center)

    @classmethod
    def generate_rect(cls, nu: int, nv: int, du: LengthQuantity, dv: LengthQuantity, center: Optional[VectorLike] = None):
        if not isinstance(dv, Q_) or not has_length_dim(dv) or not isinstance(dv.m, (int, float)):
            raise TypeError('dv must be scalar with dimension [length]')

        if not isinstance(du, Q_) or not has_length_dim(du) or not isinstance(du.m, (int, float)):
            raise TypeError('du must be scalar with dimension [length]')

        return cls(nu, nv, DimVector(du, 0*U_('µm')), DimVector(0*U_('µm'), -dv), center)

    def to_lattice(self) -> DimLattice:
        return DimLattice(self._elements)



from typing import Tuple, Sequence, Union

import numpy as np

from fibomat.rasterizedpattern import RasterizedPattern
from fibomat.shapes.rasterizedpoints import RasterizedPoints
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, scale_to, scale_factor
from fibomat.shapes import Shape, DimShape
from fibomat.mill import Mill


class PreRasterized(RasterStyle):
    def __init__(self):
        pass

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    @property
    def dimension(self) -> int:
        return 0

    @staticmethod
    def _prepare_and_scale(
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:
        # dim_shape = DimObj.create(dim_shape)

        # if isinstance(dim_shape, RasterizedPattern):
        #     rasterized_pattern = dim_shape
        #
        #     if rasterized_pattern.length_unit != out_length_unit:
        #         rasterized_pattern._dwell_points[:, :2] *= scale_factor(out_length_unit, rasterized_pattern.length_unit)
        #
        #     if rasterized_pattern.time_unit != out_time_unit:
        #         rasterized_pattern._dwell_points[:, 2] *= scale_factor(out_time_unit, rasterized_pattern.time_unit)
        #
        #     return rasterized_pattern
        if isinstance(dim_shape.shape, RasterizedPoints):
            points = np.array(dim_shape.shape.dwell_points)

            points[:, :2] *= scale_factor(out_length_unit, dim_shape.unit)
            points[:, 2] *= scale_to(out_time_unit, mill.dwell_time)

            if mill.repeats != 1:
                return RasterizedPattern(np.concatenate([points]*mill.repeats), out_length_unit, out_time_unit)
            else:
                return RasterizedPattern(points, out_length_unit, out_time_unit)
        else:
            raise TypeError('Shape must be of type RasterizedPoints or RasterizedPattern for PreRasterized style')


    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:
        return self._prepare_and_scale(dim_shape, mill, out_length_unit, out_time_unit)



from __future__ import annotations
from typing import Iterable, Any

import numpy as np

from fibomat.linalg.boundingboxes.boundingbox_base import BoundingBoxBase
from fibomat.linalg.vectors import DimVector, DimVectorLike
from fibomat.units import LengthQuantity, scale_factor, U_


class DimBoundingBox(BoundingBoxBase[DimVector, LengthQuantity]):
    _VectorClass = DimVector

    @classmethod
    def from_points(cls, points: Iterable[DimVector]) -> DimBoundingBox:
        """
        Constructs rectangular bounding box containing all `points`

        Args:
            points (VectorArrayLike): points which should be included in bounding box

        Returns:
            (BoundingBox): new `BoundingBox`
        """
        point_iter = iter(points)

        try:
            first_point = DimVector(next(point_iter))
        except StopIteration as stop_exception:
            raise ValueError('points must contain at least one point') from stop_exception

        unit = first_point.unit
        points_scaled = [first_point.vector]

        for point in point_iter:
            dim_point = DimVector(point)
            points_scaled.append((dim_point * scale_factor(unit, dim_point.unit)).vector)

        points = np.asarray(points_scaled)

        return DimBoundingBox(np.min(points, axis=0) * unit, np.max(points, axis=0) * unit)




from fibomat.linalg.transformables.transformable import Transformable
from fibomat.linalg.transformables.dim_transformable import DimTransformable
from fibomat.linalg.transformables.transformation_builder import translate, rotate, scale, mirror



"""Provide the :class:`DimensionedObj` class."""
from __future__ import annotations
from typing import Generic, TypeVar, Union, Tuple, Type

from dataclasses import dataclass

from fibomat.units import UnitType, scale_factor
from fibomat.linalg import DimVector, DimTransformable, Transformable, DimVectorLike


ObjT = TypeVar('ObjT', bound=Transformable)
UnitT = TypeVar('UnitT', bound=UnitType)
DimObjT = TypeVar('DimObjT', bound='DimensionedObj')


@dataclass(frozen=True)
class DimensionedObj(Generic[ObjT, UnitT], DimTransformable):
    """Class should be used if any kind of object must have an associated unit.

    Examples::

        dim_shape = DimensionedObj((Line([0, 0], [1, 1]), U_('µm'))

        # DimObj is an alias of DimensionedObj
        dim_shape = DimObj((Line([0, 0], [1, 1]), U_('µm'))
    """

    obj: ObjT
    unit: UnitT

    @classmethod
    def create(cls: Type[DimObjT], dim_obj: Union[DimObjT, Tuple[ObjT, UnitT]]) -> DimensionedObj[ObjT, UnitT]:
        """Parse a object and create a :class:`DimensionedObj` from a tuple or class instance.

        No checks are performed at the arguments.

        Args:
            dim_obj (DimensionedObj[[ObjT, UnitT], Tuple[ObjT, UnitT]): dimensioned object

        Returns:
            DimensionedObj[ObjT, UnitT]

        Raises:
            TypeError: Raised if `dim_obj` cannot be parsed.
        """
        if isinstance(dim_obj, cls):
            return dim_obj

        if isinstance(dim_obj, tuple):
            if len(dim_obj) != 2:
                raise ValueError('len(dim_obj) != 2. Maybe you forgot to pass a unit?')
            return cls(*dim_obj)

        raise TypeError('Cannot understand passed "dim_obj". Maybe you forgot to pass a unit?')

    @property
    def center(self) -> DimVector:
        return DimVector(self.obj.center, self.unit)

    def _impl_translate(self, trans_vec: DimVectorLike) -> None:
        trans_vec = DimVector.create(trans_vec)
        self.obj._impl_translate(scale_factor(self.unit, trans_vec.unit) * trans_vec.vector)

    def _impl_rotate(self, theta: float) -> None:
        self.obj._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        self.obj._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: DimVectorLike) -> None:
        mirror_axis = DimVector.create(mirror_axis)
        self.obj._impl_mirror(scale_factor(self.unit, mirror_axis.unit) * mirror_axis.vector)


DimObj = DimensionedObj

DimObjLike = Union[DimensionedObj[ObjT, UnitT], Tuple[ObjT, UnitT]]



"""Provides the :class:`Sample` class."""
from fibomat.linalg.boundingboxes import boundingbox
from fibomat.linalg.boundingboxes.dim_boundingbox import DimBoundingBox
from typing import Optional, List, Union, TypeVar, Type, Set
import re
import dataclasses

from fibomat.site import Site
from fibomat.linalg import DimVectorLike
from fibomat.backend import BackendBase, registry
from fibomat.utils import PathLike
from fibomat.shapes import DimShape
from fibomat.pattern import Pattern
from fibomat.layout import LayoutBase
from fibomat.default_backends import BokehBackend, StubRasterStyle
from fibomat.describable import Describable


@dataclasses.dataclass(frozen=True)
class _Annotation:
    dim_shape: DimShape
    filled: bool
    color: Optional[str]
    description: Optional[str]


BackendType = TypeVar('BackendType')


class Sample(Describable):
    """
    This class is the glueing between all subcomponents of the library.
    All shapes and their milling settings are added to this class and can be exported with the help of registered
    backends .

    """
    def __init__(self, *, description: Optional[str] = None):
        """
        Args:
            description (str, optional): Optional description of the project, default to None
        """
        super().__init__(description)

        self._sites: List[Site] = []
        self._annotations: List[_Annotation] = []

    def create_site(
            self,
            dim_position: DimVectorLike,
            dim_fov: Optional[DimVectorLike] = None,
            description: Optional[str] = None
    ) -> Site:
        """
        Creates and Site in-place (hence, the Site is automatically added to the sample). Patterns can be added to the
        returned object.

        See :class:`fibomat.site.Site.__init__` for argument description.

        Returns:
            Site
        """
        new_site: Site = Site(dim_position, dim_fov, description=description)
        self._sites.append(new_site)
        return new_site

    def add_site(self, site_like: Site) -> None:
        """
        Adds a Site to the project.
        Alternatively, the '+=' operator can be used.

        Args:
            site_like (Site): new site

        Returns:
            None
        """
        if isinstance(site_like, LayoutBase):
            for site_ in site_like.layout_elements():
                self._sites.append(site_)
        else:
            self._sites.append(site_like)

    @property
    def number_of_sites(self) -> int:
        """int: number of added sites"""
        return len(self._sites)

    @property
    def bounding_box(self) -> Optional[DimBoundingBox]:
        if not self._sites:
            return None
        else:
            bbox = self._sites[0].bounding_box

            for site in self._sites[1:]:
                bbox = bbox.extended(site.bounding_box)

            return bbox

    def __iadd__(self, site_like):
        """See :meth:`~Sample.add_site`."""
        self.add_site(site_like)
        return self

    @staticmethod
    def _export(
        backend_class: Type[BackendType], sites: Union[Site, List[Site]],
        descr_pattern: Optional[Set[str]] = None,
        **kwargs
    ) -> BackendType:
        def _matches(description_) -> bool:
            for pattern in descr_pattern:
                if re.match(pattern, description_):
                    return True
            return False

        exporter: Type[BackendBase] = backend_class(**kwargs)
        if isinstance(sites, Site):
            if descr_pattern:
                print('Warning: ignoring descriptions in _export for single site export.')
            exporter.process_site(sites)
        else:
            for site_ in sites:
                if descr_pattern:
                    description = site_.description
                    if description and _matches(description):
                        exporter.process_site(site_)
                        # descriptions.remove(description)
                else:
                    exporter.process_site(site_)

            # if descriptions:
            #    raise RuntimeError(f'Could not find sites with descriptions: {descriptions} in sample')

        return exporter

    def plot(self, show: bool = True, filename: Optional[PathLike] = None, descr_pattern: Optional[Set[str]] = None, **kwargs) -> BokehBackend:
        """
        Plots and save the project using the :class:`~fibomat.default_backends.bokeh_backend.BokehBackend`.

        Args:
            show (bool): if true, the plot is opened in a browser automatically
            filename (PathLike, optional): if filename is given, the plot is saved in this file. The file suffix should
                                           be `*.htm` or `*.html`, default to None
            `**kwargs`: parameters for the bokeh backend. These are directly passed to the __init__ method of the
                        BokehBackend class. The title parameter is automatically set to the :attr:`Sample.description`

        Returns:
            None
        """

        plotter: BokehBackend = self._export(BokehBackend, self._sites, descr_pattern=descr_pattern, title=self._description, **kwargs)

        for annot in self._annotations:
            raster = StubRasterStyle(2) if annot.filled else StubRasterStyle(1)

            plotter.process_pattern(Pattern(
                annot.dim_shape, None, raster, _annotation=True, _color=annot.color, description=annot.description
            ))

        plotter.plot()

        if filename:
            plotter.save(filename)
        if show:
            plotter.show()

        return plotter

    def export(self, exp_backend: Union[str, Type[BackendBase]], **kwargs) -> BackendBase:
        """
        Exports the project. Note that the method returns the backend object so you will be able to save a file or show
        a plot. See backends example nd docs for details.

        .. note:: The export method does not save any files on its one. This must be done by the user manually. See docs
                  of the used backend for details.

        Args:
            exp_backend (str or Type[backend.BackendBase]):
                name of the backend or class. The backend must be registered before.
            **kwargs: optional arguments are passed to the backend's __init__ method

        Returns:
            BackendBase
        """

        if isinstance(exp_backend, str):
            exp_backend = registry.get(exp_backend)

        return self._export(exp_backend, self._sites, description=self._description, **kwargs)

    def export_multi(self, exp_backend: Union[str, Type[BackendBase]], **kwargs) -> List[BackendBase]:
        """
        Similar to :meth:`Project.export` but for each :class:`fibomat.site.Site` an individual backend instance is
        returned.

        This can be usefull if multiple sites are used within fibomat but the pattern system only supports one site at a
        time.

        Returns:
            List[BackendBase]
        """
        backends: List[BackendBase] = []

        if isinstance(exp_backend, str):
            exp_backend = registry.get(exp_backend)

        for added_site in self._sites:
            backends.append(self._export(exp_backend, added_site, description=self._description, **kwargs))

        return backends

    def export_with_description(
        self, exp_backend: Union[str, Type[BackendBase]], descr_pattern: Set[str], **kwargs
    ) -> BackendBase:
        """
        Exports the project. Note that the method returns the backend object so you will be able to save a file or show
        a plot. See backends example nd docs for details.

        Only sites with a description given in `descriptions` will be exported. An exception is raised, if no site is
        found with a given description.

        .. note:: The export method does not save any files on its one. This must be done by the user manually. See docs
                  of the used backend for details.

        Args:
            exp_backend (str or Type[backend.BackendBase]):
                name of the backend or class. The backend must be registered before.
            descr_pattern (Set[str]): a set of descriptions of the sites which should be exported.
            **kwargs: optional arguments are passed to the backend's __init__ method

        Returns:
            BackendBase
        """
        if isinstance(exp_backend, str):
            exp_backend = registry.get(exp_backend)
        return self._export(exp_backend, self._sites, description=self._description, descr_pattern=descr_pattern, **kwargs)

    def add_annotation(
        self, dim_shape: DimShape, filled: bool = False, color: Optional[str] = None, description: Optional[str] = None
    ) -> None:
        """
        Add `dim_shape` to a annotation layer. This layer is only used to visualize extra shapes and is ignored by the
        exporting backend.

        Args:
            dim_shape (DimShape): shape
            filled (bool): If True, shape is plotted filled (only possible if shape is closed)
            color (str, Optional): a color bokeh can understand, default to None
            description (str, optional): description, default to None

        Returns:
            None
        """
        self._annotations.append(_Annotation(dim_shape=dim_shape, filled=filled, color=color, description=description))



"""The `mill` subpackage is used to specify the beam settings of the ion beam microscope."""
from fibomat.mill.mill import MillBase, Mill, SpecialMill
# from fibomat.mill.ionbeam import IonBeam, GaussBeam

__all__ = ['MillBase', 'Mill', 'SpecialMill']  # 'IonBeam', 'GaussBeam'



import enum

import numpy as np

from PyQt5 import QtWidgets, QtCore


from fibomat.beam_simulation.fibcanvas import FIBCanvas
from fibomat.beam_simulation.slider import Slider
from fibomat.beam_simulation.dwell_time_widget import DwellTimeWidget


@enum.unique
class AnimationState(enum.Enum):
    STARTED = enum.auto()
    STOPPED = enum.auto()
    SLIDER_STOPPED = enum.auto()


class AnimationWindow(QtWidgets.QMainWindow):
    def __init__(self, dwell_points: np.ndarray, length_unit: str, *args, **kwargs):
        super(AnimationWindow, self).__init__(*args, **kwargs)

        self._dwell_points = dwell_points

        self._update_slider_pos = True
        self._anim_state = AnimationState.STOPPED

        self._fib_canvas = FIBCanvas(dwell_points[:, :2], length_unit, self._animation_step)

        # self._dwell_time_widget = DwellTimeWidget(self._dwell_points[:, 2], self)

        self._init_gui()
        self._connect_signals()

    def _init_gui(self):
        self.setWindowTitle('Ion beam simulation')
        self.resize(800, 800)

        layout = QtWidgets.QVBoxLayout()

        self._fib_canvas.create_native()
        self._fib_canvas.native.setParent(self)
        # self._fib_canvas.native.resize(800, 600)
        self._fib_canvas.native.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        layout.addWidget(self._fib_canvas.native)

        # layout.addWidget(self._dwell_time_widget)

        layout.addWidget(QtWidgets.QLabel('Progress'))
        slider = Slider(QtCore.Qt.Horizontal)
        slider.setMinimum(0)
        slider.setMaximum(len(self._dwell_points))
        slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        slider.setTickInterval(int(len(self._dwell_points) / 10))
        layout.addWidget(slider)
        self._slider = slider

        self._start_stop_button = QtWidgets.QPushButton('Start')
        self._reset_button = QtWidgets.QPushButton('Reset')

        control_widget_layout = QtWidgets.QHBoxLayout()
        control_widget_layout.addWidget(self._start_stop_button)
        control_widget_layout.addWidget(self._reset_button)

        control_widget = QtWidgets.QWidget()
        control_widget.setLayout(control_widget_layout)
        layout.addWidget(control_widget)

        speed_label = QtWidgets.QLabel('Speed (points / s)')
        speed_label.setAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignRight)
        self._speed_input = QtWidgets.QDoubleSpinBox()
        self._speed_input.setMinimum(.1)
        self._speed_input.setMaximum(len(self._dwell_points))
        self._speed_input.setValue(50)
        self._speed_input.setSingleStep(50)

        tail_label = QtWidgets.QLabel('Tail length')
        tail_label.setAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignRight)
        self._tail_input = QtWidgets.QSpinBox()
        self._tail_input.setMinimum(1)
        self._tail_input.setMaximum(len(self._dwell_points) - 1)
        self._tail_input.setValue(100 if 100 < len(self._dwell_points) else len(self._dwell_points))
        self._tail_input.setSingleStep(50)

        alpha_label = QtWidgets.QLabel('Alpha')
        alpha_label.setAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignRight)
        self._alpha_input = QtWidgets.QDoubleSpinBox()
        self._alpha_input.setMinimum(.01)
        self._alpha_input.setMaximum(1)
        self._alpha_input.setValue(.2)
        self._alpha_input.setSingleStep(0.01)

        settings_widget_layout = QtWidgets.QHBoxLayout()
        settings_widget_layout.addWidget(speed_label)
        settings_widget_layout.addWidget(self._speed_input)
        settings_widget_layout.addWidget(tail_label)
        settings_widget_layout.addWidget(self._tail_input)
        settings_widget_layout.addWidget(alpha_label)
        settings_widget_layout.addWidget(self._alpha_input)

        settings_widget = QtWidgets.QWidget()
        settings_widget.setLayout(settings_widget_layout)
        layout.addWidget(settings_widget)

        main_widget = QtWidgets.QWidget()
        main_widget.setLayout(layout)
        self.setCentralWidget(main_widget)

    def _connect_signals(self):
        self._slider.sliderPressed.connect(self._slider_pressed)
        self._slider.sliderReleased.connect(self._slider_released)
        self._slider.valueChanged.connect(self._slider_value_changed)

        self._start_stop_button.clicked.connect(self._start_stop_clicked)
        self._reset_button.clicked.connect(self._reset_clicked)

        self._speed_input.valueChanged.connect(self._speed_value_changed)
        self._tail_input.valueChanged.connect(self._tail_value_changed)
        self._alpha_input.valueChanged.connect(self._alpha_value_changed)

        # set the default values
        self._speed_input.valueChanged.emit(self._speed_input.value())
        self._tail_input.valueChanged.emit(self._tail_input.value())
        self._alpha_input.valueChanged.emit(self._alpha_input.value())

    def _alpha_value_changed(self, alpha):
        self._fib_canvas.set_alpha(alpha)
        if self._anim_state.STOPPED:
            self._fib_canvas.update()

    def _tail_value_changed(self, length):
        self._fib_canvas.set_tail_length(length + 1)
        if self._anim_state.STOPPED:
            self._fib_canvas.update()

    def _speed_value_changed(self, speed):
        self._fib_canvas.set_speed(speed)

    def _slider_pressed(self):
        self._stop_animation()
        if self._anim_state == AnimationState.STARTED:
            self._anim_state = AnimationState.SLIDER_STOPPED

    def _slider_released(self):
        self._update_slider_pos = True

    def _slider_value_changed(self, value):
        self._fib_canvas.current_point_index = value
        if self._anim_state == AnimationState.SLIDER_STOPPED:
            self._fib_canvas.start_animation()
        else:
            self._fib_canvas.update()

    def _animation_step(self, i_step: int) -> bool:
        if i_step > self._fib_canvas.number_of_points:
            self._fib_canvas.current_point_index = self._fib_canvas.number_of_points - 1
            self._stop_animation()
            self._slider.setValue(self._fib_canvas.number_of_points - 1)

            return False
        else:
            if self._update_slider_pos:
                # self._dwell_time_widget.draw(i_step)
                self._slider.setValue(i_step)

            return True

    def _start_animation(self):
        self._anim_state = AnimationState.STARTED
        self._start_stop_button.setText('Stop')
        self._fib_canvas.start_animation()

    def _stop_animation(self):
        self._anim_state = AnimationState.STOPPED
        self._start_stop_button.setText('Start')
        self._fib_canvas.stop_animation()

    def _start_stop_clicked(self):
        if self._anim_state == AnimationState.STOPPED:
            self._start_animation()
        else:
            self._stop_animation()

    def _reset_clicked(self):
        if self._anim_state == AnimationState.STARTED:
            self._stop_animation()

        self._fib_canvas.reset_animation()
        self._slider.setValue(0)



# from fibomat.shapes.dim_shape import DimShape
from typing import List, Tuple

import numpy as np

# from fibomat.shapes import Shape, DimShape
# from fibomat.pattern import Pattern
# from fibomat.linalg import Vector


# def _have_same_type(objs: List) -> bool:
#     objs_iter = iter(objs)
#     first_elem = next(objs_iter)

#     if isinstance(first_elem, Shape):
#         base = Shape
#     elif isinstance(first_elem, DimShape):
#         base = DimShape
#     elif isinstance(first_elem, Pattern):
#         base = Pattern
#     else:
#         base = type(first_elem)

#     print(base, [isinstance(obj, base) for obj in objs_iter])
#     print(objs)

#     return all(isinstance(obj, base) for obj in objs_iter)


def _check_lattice_vectors(u, v) -> None:
    if np.isclose(abs(np.dot(np.array(u), np.array(v))), 1):
        raise ValueError('Lattice vectors may not be collinear.')


def _round_towards_mean(a: float, b: float) -> Tuple[int, int]:
    flipped = False
    if a > b:
        flipped = True
        a, b = b, a

    a = np.ceil(a)
    b = np.floor(b)

    return (b, a) if flipped else (a, b)



from typing import Union, Iterable

import numpy as np

from fibomat.linalg.vectors.vector import Vector
from fibomat.linalg.vectors.dim_vector import DimVector
from fibomat.linalg.vectors.vector_base import VectorValueError, FloatTypes

from fibomat.units import LengthQuantity


VectorLike = Union[
    Vector,
    Iterable[float]
]
"""vector-like objects for type hints"""

DimVectorLike = Union[DimVector, LengthQuantity]  # Tuple[VectorLike, LengthUnit]
"""dim-vector-like objects for type hints"""


def angle_between(vec_1: Union[VectorLike, DimVectorLike], vec_2: Union[VectorLike, DimVectorLike]) -> float:
    """Returns the the angle between two vectors.

    Args:
        vec_1 (VectorLike): first vector
        vec_2 (VectorLike): second vector

    Returns:
        float
    """
    vec_1 = np.asarray(vec_1)
    vec_2 = np.asarray(vec_2)

    if vec_1.shape != (2,) or vec_2.shape != (2,):
        raise ValueError('vec_1 and vec_2 must be VectorLike.')

    if not len(vec_1) == 2 or not len(vec_2):
        raise ValueError('vec_1 and vec_2 must have length 2.')

    return np.arccos(vec_1.dot(vec_2) / (np.linalg.norm(vec_1) * np.linalg.norm(vec_2)))


def signed_angle_between(vec_1: Union[VectorLike, DimVectorLike], vec_2: Union[VectorLike, DimVectorLike]):
    """Returns the signed the angle between two vectors.
    The angle is positive if the rotation from vec_1 to vec_2 is in positive direction
    using the smaller angle between them.

    https://stackoverflow.com/a/16544330

    Args:
        vec_1 (VectorLike): first vector
        vec_2 (VectorLike): second vector

    Returns:
        float
    """
    vec_1 = np.asarray(vec_1)
    vec_2 = np.asarray(vec_2)

    if vec_1.shape != (2,) or vec_2.shape != (2,):
        raise ValueError('vec_1 and vec_2 must be VectorLike.')

    dot = np.dot(vec_1, vec_2)
    det = np.linalg.det(np.c_[vec_1, vec_2])

    return np.arctan2(det, dot)



from __future__ import annotations
from typing import Optional, List, Iterator, Any, TypeVar, Generic
import abc

from frozenlist import FrozenList

from fibomat.layout.layoutbase import LayoutBase
from fibomat.linalg import Transformable, DimTransformable, VectorLike, DimVectorLike, BoundingBox, DimBoundingBox
# from fibomat.layout.utils import _have_same_type


ElementT = TypeVar('ElementT', Transformable, DimTransformable)
VectorT = TypeVar('VectorT', VectorLike, DimVectorLike)
BBoxT = TypeVar('BBoxT', BoundingBox, DimBoundingBox)


class GroupBase(LayoutBase[ElementT, VectorT, BBoxT], abc.ABC):
    def __init__(
        self,
        elements: List[ElementT],
        description: Optional[str] = None
    ):
        if not elements:
            raise ValueError('Length if elements must not be zero.')

        # TODO: make this works for different types (e.g. pattern, Dimgroup, ...)
        # if not _have_same_type(elements):
        #     raise TypeError('Elements in a layout must have same (base) type.')

        self._elements: FrozenList[ElementT] = FrozenList(elements)
        self._elements.freeze()

        # we evaluate this only if needed
        self._center = None
        self._bounding_box = None

        super().__init__(description=description)

    @property
    def elements(self):
        return self._elements

    def _layout_elements(self) -> Iterator[ElementT]:
        for element in self._elements:
            yield element

    @property
    def center(self) -> VectorT:
        if self._center:
            return self._center
        else:
            center = self._elements[0].center
            for element in self._elements[1:]:
                center += element.center

            center /= len(self._elements)

            self._center = center

            return center

    @property
    def bounding_box(self) -> BBoxT:
        if self._bounding_box:
            return self._bounding_box
        else:
            bbox = self._elements[0].bounding_box
            for element in self._elements[1:]:
                bbox = bbox.extended(element.bounding_box)

            self._bounding_box = bbox

            return bbox

    def _impl_translate(self, trans_vec: VectorT) -> None:
        for element in self._elements:
            element._impl_translate(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        for element in self._elements:
            element._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        for element in self._elements:
            element._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: VectorT) -> None:
        for element in self._elements:
            element._impl_mirror(mirror_axis)



"""
Backend related classes and functions.
Backends are responsible for exporting projects to a appropriate file a microscope (software) can understand.

See default_backend module for example backends.

if you Implement your own backend, make sure to register it with ::

    from fibomat.backend import registry, BackendBase

    class MyNewBackend(BackendBase):
        name = 'MyFancyName'
        ...

    registry.register(MyNewBackend, MyNewBackend.name)

The new backend can now be used within the `Project.export` method.

"""

from fibomat.backend.registry import registry
from fibomat.backend.backendbase import BackendBase

__all__ = ['registry', 'BackendBase']



from typing import Optional, Dict, Any, Callable
import collections
import datetime
import configparser

import numpy as np

from fibomat.backend import BackendBase
from fibomat import site, units, pattern, shapes, utils
from fibomat.linalg import BoundingBox
from fibomat.pattern import Pattern
from fibomat.rasterizedpattern import RasterizedPattern


def _default_save_impl(filename: utils.PathLike, dwell_points: np.ndarray, parameters: Dict[str, Any]):
    # if parameters['base_dwell_time']:
    #     dwell_points[:, 2] /= parameters['base_dwell_time']

    header = configparser.ConfigParser()

    header_dict = collections.OrderedDict()
    header_dict['length_unit'] = f'{parameters["length_unit"]:~P}'
    header_dict['time_unit'] = f'{parameters["time_unit"]:~P}'
    header_dict['fov'] = f'{parameters["fov"].width}, {parameters["fov"].height}'
    header_dict['center'] = f'{tuple(parameters["fov"].center)}'
    header_dict['base_dwell_time'] = f'{parameters["base_dwell_time"]}'
    header_dict['total_dwell_time'] = f'{parameters["total_dwell_time"]}'
    header_dict['number_of_points'] = f'{parameters["number_of_points"]}'
    header_dict['description'] = f'{parameters["description"]}'
    header_dict['time_stamp'] = f'{parameters["time_stamp"]}'

    header['Info'] = header_dict

    with open(filename, 'w') as fp:
        header.write(fp)
        fp.write('[Points]\n')
        np.savetxt(fp, dwell_points, "%.5f %.5f %d")


class SpotListBackend(BackendBase):
    name: str = 'spotlist'

    def __init__(
            self,
            #cformat_template: Optional[Tuple[str, jinja2.Template]] = None,
            save_impl: Callable = None,
            base_dwell_time: Optional[units.TimeQuantity] = None,
            length_unit: Optional[units.LengthUnit] = None,
            time_unit: Optional[units.TimeUnit] = None,
            description: Optional[str] = None,
            **kwargs
    ):
        """Spotlist backend.

        This backend rasterize all patterns and is able to output list of dwell points in the form of ::

            x1, y1, t1
            x2, y2, t2
            ...

        The exact formatting can be controlled by a custom formatting function.

        The formatting function (called save_impl in class parameters) must have the following form ::

            def func(filename: utils.PathLike, dwell_points: np.ndarray, parameters: Dict[str, Any]):
                # open the file `filename` and save dwell_points there
                ...

        The parameter dictionary holds further metadata. For all available keys, see  `_default_save_impl` in the
        sourcefile of the spotlist backend.

        Args:
            save_impl (Callable, optional): custom formatting function.
            base_dwell_time (TimeQuantity, optional):
                if given, the dwell times are autoamtically expressed as integer multiplicands of base_dwell_time.
            length_unit (LengthQuantity, optional): length unit of points.
            time_unit (TimeQuantity, optional): time unit of dwell times.
            description (str, optional): description.
            **kwargs:
        """
        super().__init__(**kwargs)

        # if format_template:
        #     if isinstance(format_template, str):
        #         self._format_template = jinja2.Template(format_template, undefined=jinja2.StrictUndefined)
        #     elif isinstance(format_template, jinja2.Template):
        #         self._format_template = format_template
        #         # self._format_template['undefined'] = jinja2.StrictUndefined
        #     else:
        #         raise TypeError('file_format must be `str` or `file_format`')
        # else:
        #     self._format_template = default_format_template

        if save_impl:
            self._save_impl = save_impl
        else:
            self._save_impl = _default_save_impl

        self._length_unit = length_unit if length_unit else units.U_('µm')
        self._time_unit = time_unit if time_unit else units.U_('µs')

        if base_dwell_time:
            self._base_dwell_time = base_dwell_time # units.scale_to(self._time_unit, base_dwell_time)
        else:
            self._base_dwell_time = None

        self._description = description

        self._fov: Optional = None
        self._site_count: int = 0
        self._site_pos: Optional = None
        self._dwell_points = []

    def process_site(self, new_site: site.Site):
        self._site_pos = new_site.center.vector_as(self._length_unit)

        self._site_count += 1

        super().process_site(new_site)

    def save(self, filename: utils.PathLike) -> None:
        if self._site_count > 1:
            print('Warning: adding merging multiple sites.')

        dwell_points = np.concatenate(self._dwell_points)

        if self._base_dwell_time:
            dwell_points[:, 2] /= units.scale_to(self._time_unit, self._base_dwell_time)

        bbox = BoundingBox.from_points(dwell_points[:, :2])

        self._save_impl(
            filename, dwell_points,
            {
                'length_unit': self._length_unit, 'time_unit': self._time_unit,
                'fov': bbox, 'base_dwell_time': self._base_dwell_time, 'total_dwell_time': np.sum(dwell_points[:, 2]),
                'number_of_points': len(dwell_points), 'description': self._description,
                'time_stamp': str(datetime.datetime.now()),
            }
        )

    def _rasterize_and_add(self, ptn: pattern.Pattern[shapes.Shape]):
        dwell_points = np.array(ptn.raster_style.rasterize(
            ptn.dim_shape, ptn.mill, self._length_unit, self._time_unit
        ).dwell_points)

        # dwell_points[:, 2] *= units.scale_to(self._time_unit, ptn.mill.dwell_time)
        dwell_points[:, :2] += self._site_pos

        self._dwell_points.append(dwell_points)

    def arc_spline(self, ptn: pattern.Pattern[shapes.ArcSpline]) -> None:
        self._rasterize_and_add(ptn)

    def rasterized_points(self, ptn: pattern.Pattern[shapes.RasterizedPoints]):
        self._rasterize_and_add(ptn)

    def rasterized_pattern(self, ptn: pattern.Pattern[RasterizedPattern]):
        self._rasterize_and_add(ptn)

    def spot(self, ptn: pattern.Pattern[shapes.Spot]) -> None:
        self._rasterize_and_add(ptn)

    def line(self, ptn: pattern.Pattern[shapes.Line]) -> None:
        self._rasterize_and_add(ptn)

    def polyline(self, ptn: pattern.Pattern[shapes.Polyline]) -> None:
        self._rasterize_and_add(ptn)

    def polygon(self, ptn: pattern.Pattern[shapes.Polygon]) -> None:
        self._rasterize_and_add(ptn)

    def arc(self, ptn: pattern.Pattern[shapes.Arc]) -> None:
        self._rasterize_and_add(ptn)

    def circle(self, ptn: pattern.Pattern[shapes.Circle]) -> None:
        self._rasterize_and_add(ptn)

    def ellipse(self, ptn: pattern.Pattern[shapes.Ellipse]) -> None:
        self._rasterize_and_add(ptn)

    def rect(self, ptn: pattern.Pattern[shapes.Rect]) -> None:
        self._rasterize_and_add(ptn)







from fibomat.shapes.shape import Shape
from fibomat.shapes.dim_shape import DimShape
from fibomat.shapes.arc_spline import ArcSpline, ArcSplineCompatible
from fibomat.shapes.line import Line
from fibomat.shapes.polyline import Polyline
from fibomat.shapes.polygon import Polygon
from fibomat.shapes.spot import Spot
from fibomat.shapes.rect import Rect
from fibomat.shapes.circle import Circle
from fibomat.shapes.ellipse import Ellipse
from fibomat.shapes.parametric_curve import ParametricCurve
from fibomat.shapes.arc import Arc
from fibomat.shapes.rasterizedpoints import RasterizedPoints
from fibomat.shapes.hollow_arc_spline import HollowArcSpline
from fibomat.shapes.biarc import Biarc
from fibomat.shapes.text import Text


__all__ = [
    'Shape', 'ArcSpline', 'Line', 'Polyline', 'Polygon', 'Spot', 'Rect', 'Circle', 'Ellipse', 'ParametricCurve', 'Arc',
    'RasterizedPoints', 'HollowArcSpline', 'DimShape', 'ArcSplineCompatible', 'Biarc', 'Text'
]



"""Provide the :class:`LayoutBase` class."""
from __future__ import annotations
from typing import Optional, Iterator, TypeVar, Generic
import abc

from fibomat.linalg import (
    Transformable, DimTransformable, VectorLike, DimVectorLike, BoundingBox, DimBoundingBox
)


ElementT = TypeVar('ElementT', Transformable, DimTransformable)
VectorT = TypeVar('VectorT', VectorLike, DimVectorLike)
BBoxT = TypeVar('BBoxT', BoundingBox, DimBoundingBox)


class LayoutBase(Generic[ElementT, VectorT, BBoxT], abc.ABC):
    """
    ABC for all layout related classes.

    It can be used to arrange :class:`fibomat.site.Site`, :class:`fibomat.pattern.Pattern` and
    :class:`fibomat.shapes.Shape`.

    The saved are accessed via the :meth:`LayoutBase.layout_elements` method which returns a generator containing all
    included elements.

    What kind of elements and how these are set must be specified in child classes.

    """

    def __init__(self, description: Optional[str] = None):
        """
        Args:
            description (str, optional): description
        """
        super().__init__(description=description)

    @abc.abstractmethod
    def _layout_elements(self) -> Iterator[ElementT]:
        raise NotImplementedError

    def layout_elements(self) -> Iterator[ElementT]:
        """Access to the saved elements.

        Yields:
            Any: Type depends on saved element.
        """
        for element in self._layout_elements():
            if isinstance(element, LayoutBase):
                for sub_element in element._layout_elements():  # pylint: disable=protected-access
                    yield sub_element
            else:
                yield element

    # @property
    # def pivot(self):
    #     if self._pivot is not None:
    #         if self._is_dimensioned:
    #             return DimVector.create(self._pivot(self))
    #         else:
    #             return Vector(self._pivot(self))
    #     return self.center
    #
    # @pivot.setter
    # def pivot(self, value):
    #     self._pivot = value
    #
    # @property
    # def bounding_box(self) -> boundingbox.BoundingBox:
    #     """Bounding box of shape (getter)
    #
    #     Access:
    #         get
    #
    #     Returns:
    #         BoundingBox
    #
    #     Raises:
    #         RuntimeError: Raised if Layout object does not contain any elements
    #     """
    #     raise NotImplementedError
    #
    #     element_iter = self.layout_elements()
    #
    #     try:
    #         bbox = next(element_iter).bounding_box
    #     except StopIteration:
    #         raise RuntimeError('Cannot calculate bounding box of empty Layout.')  # pylint: disable=raise-missing-from
    #
    #     for element in element_iter:
    #         bbox = bbox.extended(element.bounding_box)
    #
    #     return bbox



"""Provide some utility functions."""

import numpy as np

from fibomat.utils.pathlike import PathLike
from fibomat.utils.math import float_gcd, mod_2pi


def make_read_only_view(array: np.ndarray):
    view = array


__all__ = ['PathLike', 'float_gcd', 'mod_2pi']




import numpy as np

from PyQt5 import QtWidgets

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt


# https://github.com/vispy/vispy/blob/af676d184be4aa778676963f054946d12e1b3ef9/examples/basics/scene/isocurve_for_trisurface_qt.py
class DwellTimeWidget(QtWidgets.QWidget):
    def __init__(self, dwell_times, parent):
        super().__init__(parent)

        self._dwell_times = dwell_times

        self.figure = plt.figure()
        self.figure.clear()

        ax = self.figure.add_subplot(111)

        ax.set_xlim(0, len(dwell_times)-1)
        ax.set_ylim(min(dwell_times), max(dwell_times))
        ax.set_xticks([])
        ax.tick_params(axis="y", direction="in", pad=-40)

        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_color('lightgrey')

        ax.yaxis.grid(color='lightgrey')

        plt.subplots_adjust(left=0, right=1, wspace=0, hspace=0) # bottom=0.0, right=1, top=1,

        self.dwell_time_canvas = FigureCanvas(self.figure)
        self.dwell_time_canvas.setGeometry(self.rect())

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.dwell_time_canvas)
        self.setLayout(layout)
        self.setFixedHeight(200)

        # ax.plot(dwell_times, '-', color='black')
        self.current_point, = ax.plot([], [], 'ro')


        self.dwell_time_canvas.draw()

        self.axbackground = self.figure.canvas.copy_from_bbox(ax.bbox)

        self.ax = ax

        # plt.show(block=False)

        # refresh canvas

    def draw(self, index: int):
        self.current_point.set_xdata([index])
        self.current_point.set_ydata([self._dwell_times])

        self.figure.canvas.restore_region(self.axbackground)
        self.ax.draw_artist(self.current_point)
        self.figure.canvas.blit(self.ax.bbox)

        self.figure.canvas.flush_events()

        #self.dwell_time_canvas.draw()

    def start_animation(self):
        # self.timer.start()
        self._run_animation = True
        self._last_update = self._timer.elapsed

    def stop_animation(self):
        # self.timer.stop()
        self._run_animation = False

    def reset_animation(self):
        self._last_update = 0
        self._current_point = 0
        self._run_animation = False



"""
This submodule provides :class:`~fibomat.linalg.linalg.Vector` and :class:`~fibomat.linalg.vectorarray.VectorArray`
classes with some utility functions.

Example ::

    import numpy as np
    from fibomat import linalg

    v = linalg.Vector(1, 2)
    u = linalg.Vector(r=1, phi=np.pi/3)  # polar form

    # access to elements
    v_x = v.x
    # or
    v_x = v[0]

    v_r = v.r
    v_phi = v.phi

    # changing the linalg
    v_x = 1
    v_phi = 2 * np.phi/3

    v.rotate(np.pi/4, center=(1, 1)).scale((2, 2)).translate((-1, 2))

    # use it for calculations
    w = v + w
    w *= 2

    # ...

Nearly the same can be done with the :class:`~fibomat.linalg.vectorarray.VectorArray` class, which is a list of vectors.
Example ::

    import numpy as np
    from fibomat import linalg

    va = linalg.VectorArray((1, 1), (2, 2), (3, 3))
    # or
    va = linalg.VectorArray([(1, 1), (2, 2), (3, 3)])
    # or
    va = linalg.VectorArray(np.array([(1, 1), (2, 2), (3, 3)]))

    # access to elements
    va_x = va.x  # list of x values

    v = va[0]  # returns the first linalg
    v_x = va[0].x  # or v_x = va[0][0]

    # changing the linalg array
    # all operations are applied to each linalg of the list

    va.rotated(np.pi/4, center=(1, 1)).scale((2, 2)).translate((-1, 2))
"""

from fibomat.linalg.vectors import (
    Vector, VectorLike, DimVector, DimVectorLike, VectorValueError, angle_between, signed_angle_between
)
# from fibomat.linalg.vectorarray import VectorArray, VectorArrayLike, VectorArrayValueError
from fibomat.linalg.transformables.transformable import Transformable
from fibomat.linalg.transformables.dim_transformable import DimTransformable
from fibomat.linalg.transformables.transformation_builder import translate, rotate, scale, mirror
# from fibomat.linalg.dimvector import DimVector, DimVectorLike
from fibomat.linalg.transformables.transformable_base import TransformableBase
from fibomat.linalg.boundingboxes import BoundingBox, DimBoundingBox


__all__ = [
    'Vector', 'VectorLike', 'DimVector', 'DimVectorLike', 'VectorValueError', 'angle_between', 'signed_angle_between',
    'Transformable', 'DimTransformable', 'TransformableBase', 'translate', 'rotate', 'mirror', 'scale',
    'BoundingBox', 'DimBoundingBox'
]



"""
The layout submodule provides tools to arrange :class:`fibomat.site.Site`, :class:`fibomat.pattern.Pattern` and
:class:`fibomat.shapes.Shape`.
"""
from fibomat.layout.layoutbase import LayoutBase
from fibomat.layout.groups.group import Group
from fibomat.layout.groups.dim_group import DimGroup
from fibomat.layout.lattices.lattice import Lattice
from fibomat.layout.lattices.dim_lattice import DimLattice
from fibomat.layout.lattices.lattice_builder import LatticeBuilder
from fibomat.layout.lattices.dim_lattice_builder import DimLatticeBuilder

__all__ = ['LayoutBase', 'Lattice', 'DimLattice', 'Group', 'DimGroup', 'LatticeBuilder', 'DimLatticeBuilder']



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

from fibomat.sample import Sample
from fibomat.site import Site
from fibomat.pattern import Pattern
from fibomat.mill import Mill
from fibomat.linalg import Vector, DimVector
from fibomat.units import U_, Q_

import fibomat.default_backends

__version__ = '0.3.9'

# try:
#     # Change here if project is renamed and does not equal the package name
#     dist_name = __name__
#     __version__ = get_distribution(dist_name).version
#     """package version"""
# except DistributionNotFound:
#     __version__ = 'unknown'
# finally:
#     del get_distribution, DistributionNotFound

__all__ = ['__version__', 'Sample', 'Site', 'Pattern', 'Mill', 'Vector', 'U_', 'Q_']



from typing import Optional, Union, List

import numpy as np

from fibomat.linalg import VectorLike, Vector, Transformable
from fibomat.linalg.transformables.transformable_base import VectorT, BBoxT
from fibomat.shapes.arc import Arc, ArcSpline, ArcSplineCompatible
from fibomat.shapes import Line


class Biarc(Transformable, ArcSplineCompatible):
    """
    https://www.ryanjuckett.com/biarc-interpolation/
    """

    @classmethod
    def _make_biarc_seg(cls, start: Vector, end: Vector, tangent: Vector, reversed: bool) -> Union[Arc, Line, None]:
        """Create a arc or line from ``start`` to ``end`` and tangent ``tangent`` at start.

        Args:
            start (Vector): start
            end (Vector):
            tangent (Vector): tangent at start
            reversed (bool): if True, the arc is reversed

        Returns:
            Union[Arc, Line, None]: None is returned if start == end, Line is returned if (end - start) || tangent and
                                    Arc otherwise
        """
        # https://www.ryanjuckett.com/biarc-interpolation/
        tangent = tangent.normalized()
        normal = Vector(-tangent.y, tangent.x)
        end_to_start = end - start

        if np.allclose(end_to_start, 0.):
            # raise cls._ZeroArcAngle('arc angle ist zero.')
            return None

        u_denominator = 2 * normal.dot(end_to_start)

        if np.isclose(u_denominator, 0.):
            if reversed:
                return Line(end, start)
            else:
                return Line(start, end)

        u = end_to_start.dot(end_to_start) / u_denominator

        center = start + u * normal
        radius = abs(u)

        center_to_start = (start - center) / radius
        center_to_end = (end - center) / radius

        sweep = center_to_start.cross(center_to_end)

        delta_angle = np.sign(sweep) * np.arccos(center_to_start.dot(center_to_end))

        start_angle = (start - center).angle_about_x_axis
        end_angle = start_angle + delta_angle

        if reversed:
            return Arc(
                radius=radius,
                start_angle=end_angle, end_angle=start_angle,
                sweep_dir=not(sweep > 0),
                center=center
            )
        else:
            return Arc(
                radius=radius,
                start_angle=start_angle, end_angle=end_angle,
                sweep_dir=sweep > 0,
                center=center
            )

    @classmethod
    def _make_biarc_segs(
        cls, p_1: Vector, p_2: Vector, t_1: Vector, t_2: Vector, d: float
    ) -> List[Union[Arc, Line, None]]:
        """Create arcs of biarc for given d value. See source for definition of ``d``.

        Args:
            p_1 (Vector): start of biarc
            p_2 (Vector): end of biarc
            t_1 (Vector): tangent at start
            t_2 (Vector): tangent at end
            d (float): d value

        Returns:
            Tuple[Arc, Arc]
        """
        midpoint = (p_1 + p_2 + d * (t_1 - t_2)) / 2

        seg_1 = cls._make_biarc_seg(p_1, midpoint, t_1, False)
        seg_2 = cls._make_biarc_seg(p_2, midpoint, t_2, True)

        return [seg for seg in [seg_1, seg_2] if seg]

    def __init__(
        self,
        p_1: VectorLike, p_2: VectorLike,
        t_1: VectorLike, t_2: VectorLike,
        description: Optional[str] = None
    ):
        """Interpolate a biarc to given points and tangents.

        For interpolation details see references.

        References:
            - https://www.ryanjuckett.com/biarc-interpolation/
            - https://ieeexplore.ieee.org/document/5390085

        Args:
            p_1 (Vector): start of biarc
            p_2 (Vector): end of biarc
            t_1 (Vector): tangent at start
            t_2 (Vector): tangent at end
            description (str, optional): optional description
        """
        super().__init__(description)

        p_1 = Vector(p_1)
        p_2 = Vector(p_2)

        t_1 = Vector(t_1).normalized()
        t_2 = Vector(t_2).normalized()

        t = t_1 + t_2
        v = p_2 - p_1

        if np.allclose(v, 0.):
            raise ValueError('p_1 == p_2. No biarc can be interpolated.')

        d_denominator = 2. * (1. - t_1.dot(t_2))

        if np.isclose(d_denominator, 0.):
            if np.isclose(v.dot(t_2), 0.):
                # create to semicircles
                midpoint = p_1 + v / 2
                c_1 = p_1 + .25 * v
                c_2 = p_1 + .75 * v

                sweep = v.cross(t_2)

                segments = [
                    Arc.from_points_center(p_1, midpoint, c_1, sweep < 0),
                    Arc.from_points_center(midpoint, p_2, c_2, sweep > 0)
                ]
            else:
                d = v.dot(v) / (4. * v.dot(t_2))
                segments = self._make_biarc_segs(p_1, p_2, t_1, t_2, d)
        else:
            d_discriminant = v.dot(t) * v.dot(t) + d_denominator * v.dot(v)
            d = (-v.dot(t) + np.sqrt(d_discriminant)) / d_denominator

            segments = self._make_biarc_segs(p_1, p_2, t_1, t_2, d)

        assert 1 <= len(segments) <= 2
        self._biarc = ArcSpline.from_segments(segments)

    @property
    def segments(self) -> List[Union[Arc, Line]]:
        return self._biarc.segments

    def to_arc_spline(self) -> ArcSpline:
        return self._biarc

    @property
    def center(self) -> VectorT:
        """Center is mean value of start of first segment and end of last segment (hence, it is compatible with ArSpline)

        Returns:

        """
        return self._biarc.center

    @property
    def bounding_box(self) -> BBoxT:
        return self._biarc.bounding_box

    def _impl_translate(self, trans_vec: VectorT) -> None:
        self._biarc._impl_translate(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        self._biarc._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        self._biarc._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: VectorT) -> None:
        self._biarc._impl_mirror(mirror_axis)

# from fibomat import Sample, U_
# from fibomat.shapes import Spot
#
# s = Sample()
#
# # two arcs
# # biarc = Biarc(Vector(0, 1), Vector(3, 0), Vector(-1, 0), Vector(0, 1))
#
# # only one arc
# # biarc = Biarc(Vector(-1, 0), Vector(1, 0), Vector(1, -1), Vector(-1, -1))
#
# # line + arc
# # biarc = Biarc(Vector(0, 0), Vector(3, -1), Vector(1, 0), Vector(0, -1))
#
# # line + line
# # biarc = Biarc(Vector(0, 0), Vector(3, 0), Vector(1, 0), Vector(1, 0))
#
# # two arcs (degenerate case, two semicircles)
# # biarc = Biarc(Vector(0, 1), Vector(0, -1), Vector(1, 0), Vector(1, 0))
#
# # two arcs (degenerate case, but no semicircles)
# biarc = Biarc(Vector(0, 1), Vector(1, -1), Vector(1, 0), Vector(1, 0))
#
#
# # print()
# #
# # arc = arc_center_radius_from_start_end_tangent(Vector(100, 100), Vector(200, 0), Vector(-1, 0))
# # print(arc)
#
#
# # s.add_annotation(arc * U_('µm'))
#
# for a in biarc.segments:
#     print(a.start, a.end)
#     s.add_annotation(a * U_('µm'))
#
# # s.add_annotation(Arc.from(Vector(100, 100), Vector(200, 0), Vector(-1, 0)) * U_('µm'))
#
# s.plot()



"""Module provides support for import of vector graphics."""
from typing import Optional
import pathlib

from fibomat.shapes import Shape

from fibomat.from_file.dxf import shapes_from_dxf
from fibomat.from_file.svg import shapes_from_svg
from fibomat.utils import PathLike
from fibomat.layout import Group
from fibomat.units import LengthQuantity


def shapes_from_file(file_path: PathLike, scale: LengthQuantity) -> Optional[Group]:
    """Parse a vector graphic file and return the contained data mapped to fib-o-mat shape types.

    The file type is determined by the file_path suffix. Currently supported is only dfx-format.

    Args:
        file_path (PathLike): file to be parsed
        scale (LengthQuantity): scale to be used for import.  E.g.: if scale = 1cm, 1cm in a svg file will be 1 length unit in fibomat.

    Returns:
        List[Shapes]

    Raises:
        RuntimeError: Raised if provided file is not valid.
        RuntimeError: Raised if provided file format is not supported.
    """
    file_path = pathlib.Path(file_path).absolute()

    if not file_path.is_file():
        raise RuntimeError(f'"{file_path}" is not a valid file')

    suffix = file_path.suffix

    if suffix == '.dxf':
        return Group(shapes_from_dxf(file_path))
    elif suffix == '.svg':
        return shapes_from_svg(file_path, scale)

    raise RuntimeError(f'Cannot import file with extension "{suffix}"')



"""Provide the :class:`Mill` class."""
from typing import Optional

from fibomat.units import QuantityType, has_time_dim, Q_
from fibomat.mill.ionbeam import IonBeam


class MillBase:
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def __repr__(self) -> str:
        return ('{}(' + ', '.join([key + '={}' for key in self._kwargs.keys()]) + ')').format(
            self.__class__.__name__, *self._kwargs.values()
        )

    # def __getattr__(self, item):
    #     return self.__getitem__(item)
    #
    # def __setattr__(self, key, value):
    #     if not key.startswith('_'):
    #         raise TypeError(
    #             "'MillBase' object does not support item assignment (if you are a developer, use variables with "
    #             " _ [underscore] as first character. Then, this check is bypassed)"
    #         )
    #     super().__setattr__(key, value)

    # def __getattribute__(self, item: str):
    #     print('__getattribute__', item)
    #     if not item.startswith('_') and item not in self._kwargs:
    #         raise AttributeError
    #     return object.__getattribute__(self, item)

        #
    # def __setattr__(self, key, value):
    #     raise NotImplementedError

    def __getitem__(self, item: str):
        try:
            return self._kwargs[item]
        except KeyError as key_error:
            raise KeyError(
                f'Mill object does not have property "{item}". '
                'Maybe you need a SpecialMill with custom properties for the used backend?'
            ) from key_error


class Mill(MillBase):
    """The `Mill` class is used to specify the, dwell_time per spot and the number of repats for a pattern.

    Optionally, the class can hold an object describing the shape of the ion beam which is needed if any kind of
    optimization is done.
    """
    def __init__(self, dwell_time: QuantityType, repeats: int):
        """
        Args:
            dwell_time (QuantityType): dwell time per spot
            repeats (int): number of repeats
        """

        if not isinstance(dwell_time, Q_):
            raise TypeError('dwell_time must be a quantity.')

        if not has_time_dim(dwell_time):
            raise ValueError('dwell_time must have dimension [time].')

        if not isinstance(repeats, int):
            raise TypeError('repats must be an int')

        if repeats < 1:
            raise ValueError('repeats must be at least 1.')

        super().__init__(dwell_time=dwell_time, repeats=repeats)

    @property
    def dwell_time(self) -> QuantityType:
        return self['dwell_time']

    @property
    def repeats(self) -> int:
        return self['repeats']


class SpecialMill(MillBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)




from typing import Optional, Sequence, List
from collections import deque

import numpy as np

from fibomat.linalg import VectorLike, BoundingBox, Vector
from fibomat.shapes.shape import Shape
from fibomat.shapes.arc_spline import ArcSpline

from fibomat.curve_tools import combine_curves


class HollowArcSpline(Shape):
    def __init__(
        self,
        boundary: ArcSpline,
        holes: Optional[Sequence[ArcSpline]] = None,
        description: Optional[str] = None
    ):
        super().__init__(description)

        if not isinstance(boundary, ArcSpline):
            raise TypeError('boundary must be an ArcSplines.')

        if not all(isinstance(hole, ArcSpline) for hole in holes):
            raise TypeError('holes must be ArcSplines.')

        if not boundary.is_closed:
            raise ValueError('boundary must be closed arc spline.')

        if not holes:
            holes = []

        if not all(hole.is_closed for hole in holes):
            raise ValueError('holes must be closed arc splines')

        self._boundary, self._holes = self._exclude_holes(boundary, self._merge_holes(holes))

    @staticmethod
    def _exclude_holes(boundary: ArcSpline, holes: Sequence[ArcSpline]):
        queue = deque(holes)

        non_intersecting_holes = []

        while queue:
            hole = queue.popleft()

            # if not combine_curves(boundary, hole, mode='intersect'):
            #     raise RuntimeError('Hole is not included in boundary.')

            excluded = combine_curves(boundary, hole, mode='exclude')

            if not excluded['remaining']:
                # something strange happened
                raise RuntimeError
            if len(excluded['remaining']) == 1:
                if len(excluded['remaining'][0].vertices) == len(boundary.vertices) and np.allclose(excluded['remaining'][0].vertices, boundary.vertices):
                    non_intersecting_holes.append(hole)
                else:
                    boundary = excluded['remaining'][0]
            if len(excluded['remaining']) > 1:
                raise RuntimeError(
                    'Shape is not simply connected.'
                    'This is most likely caused by a hole cutting the shape in two or more pieces.'
                )

        return boundary, non_intersecting_holes

    @staticmethod
    def _merge_holes(holes: Sequence[ArcSpline]):
        if holes:
            queue = deque(holes)

            non_intersection_holes = []

            while queue:
                hole = queue.popleft()
                non_intersecting = True

                for i in range(len(queue)):
                    if hole.bounding_box.overlaps_with(queue[i].bounding_box):
                        union = combine_curves(hole, queue[i], mode='union')

                        if union['subtracted']:
                            raise RuntimeError(
                                'Shape is not simply connected. '
                                'This is most likely caused by holes which separate the shape in two or more parts.'
                            )

                        if len(union['remaining']) == 1:
                            del queue[i]
                            queue.append(union['remaining'][0])
                            non_intersecting = False
                            break

                if non_intersecting:
                    non_intersection_holes.append(hole)

            return non_intersection_holes
        else:
            return []

    @property
    def boundary(self) -> ArcSpline:
        return self._boundary

    @property
    def holes(self) -> List[ArcSpline]:
        return self._holes

    def __repr__(self) -> str:
        return self.__class__.__name__

    @property
    def is_closed(self) -> bool:
        return True

    def contains(self, pos: VectorLike) -> bool:
        """Returns True if pos is contained in shape

        Args:
            pos (VectorLike): point to be tested

        Returns:
            bool
        """
        return self._boundary.contains(pos) and not any([hole.contains(pos) for hole in self._holes])

    @property
    def center(self) -> Vector:
        return self._boundary.center

    @property
    def bounding_box(self) -> BoundingBox:
        return self._boundary.bounding_box

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._boundary._impl_translate(trans_vec)

        for hole in self._holes:
            hole._impl_translate(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        self._boundary._impl_rotate(theta)

        for hole in self._holes:
            hole._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        self._boundary._impl_scale(fac)

        for hole in self._holes:
            hole._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._boundary._impl_mirror(mirror_axis)

        for hole in self._holes:
            hole._impl_mirror(mirror_axis)






import numpy as np

from vispy.visuals import Visual
from vispy.gloo import VertexBuffer


class Tail(Visual):
    VERTEX_SHADER = """
        #version 120
        uniform float u_size;

        attribute vec2 a_position;
        attribute vec4 a_color;

        varying vec4 v_fg_color;
        varying vec4 v_bg_color;
        varying float v_radius;
        varying float v_linewidth;
        varying float v_antialias;
        void main (void) {
            v_radius = u_size;
            v_linewidth = 1.0;
            v_antialias = 1.0;
            v_fg_color  = vec4(0.0,0.0,0.0,0.1);
            v_bg_color  = a_color;
            gl_Position = $transform(vec4(a_position,0,1));
            gl_PointSize = 2.0*(v_radius + v_linewidth + 1.5*v_antialias);
        }
    """

    FRAGMENT_SHADER = """
        #version 120
        varying vec4 v_fg_color;
        varying vec4 v_bg_color;
        varying float v_radius;
        varying float v_linewidth;
        varying float v_antialias;
        void main()
        {
            float size = 2.0*(v_radius + v_linewidth + 1.5*v_antialias);
            float t = v_linewidth/2.0-v_antialias;
            float r = length((gl_PointCoord.xy - vec2(0.5,0.5))*size);
            float d = abs(r - v_radius) - t;
            if( d < 0.0 )
                gl_FragColor = v_fg_color;
            else
            {
                float alpha = d/v_antialias;
                alpha = exp(-alpha*alpha);
                if (r > v_radius)
                    gl_FragColor = vec4(v_fg_color.rgb, alpha*v_fg_color.a);
                else
                    gl_FragColor = mix(v_bg_color, v_fg_color, alpha);
            }
        }
    """

    def __init__(self, dwell_points: np.ndarray, tail_length: int, size: float):
        super().__init__(self.VERTEX_SHADER, self.FRAGMENT_SHADER)

        self.set_gl_state(blend=True, blend_func=('src_alpha', 'one_minus_src_alpha'))
        self._draw_mode = 'points'

        self._tail_length = tail_length
        self._size = size
        self._alpha = 0.5

        self._dwell_points = dwell_points

    def _prepare_transforms(self, view=None):
        view.view_program.vert['transform'] = view.transforms.get_transform()

    def _prepare_draw(self, view):
        # print('prepare')
        # attributes / uniforms are not available until program is built
        pass

    def draw(self, current_point: int):
        if current_point > self._tail_length:
            i_min = current_point - self._tail_length
            i_max = current_point
        else:
            i_min = 0
            i_max = current_point

        n = i_max - i_min

        v_color = np.c_[np.full(n, 1.), np.zeros(n), np.zeros(n), np.linspace(0, self._alpha, n)].astype(np.float32)

        self.shared_program['a_position'] = VertexBuffer(self._dwell_points[i_min:i_max, :2])
        self.shared_program['a_color'] = VertexBuffer(v_color)
        self.shared_program['u_size'] = self._size

        super().draw()


class DwellPoints(Visual):
    VERTEX_SHADER = """
        #version 120
        uniform float u_size;
        uniform vec4 u_color;
        uniform float u_alpha_border;

        attribute vec2 a_position;

        varying vec4 v_fg_color;
        varying vec4 v_bg_color;
        varying float v_radius;
        varying float v_linewidth;
        varying float v_antialias;
        void main (void) {
            v_radius = u_size;
            v_linewidth = 1.0;
            v_antialias = 1.0;
            v_fg_color  = vec4(0.0,0.0,0.0,u_alpha_border);
            v_bg_color  = u_color;
            gl_Position = $transform(vec4(a_position,0,1));
            gl_PointSize = 2.0*(v_radius + v_linewidth + 1.5*v_antialias);
        }
    """

    FRAGMENT_SHADER = """
        #version 120
        varying vec4 v_fg_color;
        varying vec4 v_bg_color;
        varying float v_radius;
        varying float v_linewidth;
        varying float v_antialias;
        void main()
        {
            float size = 2.0*(v_radius + v_linewidth + 1.5*v_antialias);
            float t = v_linewidth/2.0-v_antialias;
            float r = length((gl_PointCoord.xy - vec2(0.5,0.5))*size);
            float d = abs(r - v_radius) - t;
            if( d < 0.0 )
                gl_FragColor = v_fg_color;
            else
            {
                float alpha = d/v_antialias;
                alpha = exp(-alpha*alpha);
                if (r > v_radius)
                    gl_FragColor = vec4(v_fg_color.rgb, alpha*v_fg_color.a);
                else
                    gl_FragColor = mix(v_bg_color, v_fg_color, alpha);
            }
        }
    """

    def __init__(self, dwell_points: np.ndarray):
        super().__init__(self.VERTEX_SHADER, self.FRAGMENT_SHADER)
        self.set_gl_state(blend=True, blend_func=('src_alpha', 'one_minus_src_alpha'))
        self._draw_mode = 'points'

        self._dwell_points = dwell_points

        self.shared_program['a_position'] = VertexBuffer(dwell_points)
        self.shared_program['u_color'] = (0, 0, 0, 1)
        self.shared_program['u_size'] = 1
        self.shared_program['u_alpha_border'] = 1

    def _prepare_transforms(self, view=None):
        view.view_program.vert['transform'] = view.transforms.get_transform()

    def _prepare_draw(self, view):
        pass



"""
Provides the :class:`TransformableBase` class.
"""
from __future__ import annotations
from typing import Optional, Union, TypeVar, Literal, Type, Generic, Callable
import abc

import numpy as np

from fibomat.linalg.vectors import Vector, VectorLike, DimVector
from fibomat.describable import Describable
from fibomat.linalg.transformables.transformation_builder import (
    _TransformationBuilder, _TranslationBuilder, _RotationBuilder, _ScaleBuilder, _MirrorBuilder
)
from fibomat.linalg.boundingboxes import BoundingBox, DimBoundingBox
from fibomat.units import LengthQuantity


VectorT = TypeVar('VectorT', Vector, DimVector)
ScalarT = TypeVar('ScalarT', int, float, np.integer, np.float, LengthQuantity)
BBoxT = TypeVar('BBoxT', BoundingBox, DimBoundingBox)
SelfT = TypeVar('SelfT', bound='TransformableBase')


class TransformableBase(Describable, Generic[VectorT, ScalarT, BBoxT], abc.ABC):
    """
    :class:`Transformable` is a base class providing the translate, rotate and uniform scale
    transformations.

    In order to use this mixin in a child class, the following methods and properties must be implemented:
        * :attr:`Transformable.center`
        * :meth:`Transformable.translate`
        * :meth:`Transformable.simple_rotate`
        * :meth:`Transformable.simple_scale`

    """

    _VectorClass: Type[VectorT]

    def __init__(self: SelfT, description: Optional[str] = None):
        """
        Args:
            pivot (VectorLike, optional): if set, the :attr:`Transformable.pivot` is set to `pivot`. If not set,
                                           :attr:`Transformable.center` is used as default.

            description (str, optional): optional description
        """
        super().__init__(description)

        self._pivot: Optional[Callable[[SelfT], VectorT]] = None

    @property
    @abc.abstractmethod
    def center(self) -> VectorT:
        """center of the (geometric) object

        Access:
            get

        Returns:
            Any
        """
        raise NotImplementedError

    @property
    def pivot(self) -> VectorT:
        """Origin of the (geometric) object. If origin is set to `None`, :attr:`Transformable.center` will be
        returned.

        Pivot must be set to a callable function without parameters. ::

            transformable_obj = ...
            transformable_obj.pivot = lambda: return Vector(1, 2)
            print(transformable_obj.pivot)  # will print Vector(1, 2)

        Access:
            get/set

        Returns:
            Vector
        """
        # if self._pivot is not None:
        #     return Vector(self._pivot(self))
        # return self.center
        if self._pivot is not None:
            return self._VectorClass(self._pivot(self))
        return self.center

    @pivot.setter
    def pivot(self: SelfT, value: Callable[[SelfT], VectorLike]):
        self._pivot = value

    @property
    @abc.abstractmethod
    def bounding_box(self) -> BBoxT:
        """
        :class:`~fibomat.boundingbox.BoundingBox`: bounding box of transformable

        Access:
            get
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _impl_translate(self, trans_vec: VectorT) -> None:
        """Translate the object by `trans_vec`.

        Args:
            trans_vec (VectorLike): translation linalg

        Returns:
            None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _impl_rotate(self, theta: float) -> None:
        """Rotate the object by theta.

        Args:
            theta: rotation angle

        Returns:
            None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _impl_scale(self, fac: float) -> None:
        """Scale the object by fac.

        Args:
            fac: scale factor

        Returns:
            None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _impl_mirror(self, mirror_axis: VectorT) -> None:
        """Mirror the object at mirror_axis.

        Args:
            mirror_axis (VectorLike): mirror_plane

        Returns:
            None
        """
        raise NotImplementedError

    def _apply_shifted_trafo(
        self: SelfT,
        trafo: Callable[[float], None],
        arg: float,
        origin: Optional[Union[VectorT, str]] = None
    ) -> SelfT:
        if origin:
            if isinstance(origin, str):
                if origin == 'center':
                    origin = self.center
                elif origin == 'pivot':
                    origin = self.pivot
                else:
                    raise ValueError(f'Unknown origin `{origin}`')
            else:
                origin = self._VectorClass(origin)

            self._impl_translate(-origin)  # pylint: disable=invalid-unary-operand-type
            trafo(float(arg))
            self._impl_translate(origin)
        else:
            trafo(float(arg))

        return self

    def translated_to(self: SelfT, pos: VectorT) -> SelfT:
        """Return a translated copy of the object so that self.pivot == pos

        Args:
            pos: new position of object

        Returns:
            TransformableBase
        """
        return self.translated(self._VectorClass(pos) - self.pivot)

    def translated(self: SelfT, trans_vec: VectorT) -> SelfT:
        """Return a translated copy of the object by `trans_vec`.

        Args:
            trans_vec (VectorLike): translation vector

        Returns:
            TransformableBase
        """
        # pylint: disable=protected-access

        # if not np.isclose(float(self._VectorClass(trans_vec).mag), 0.):
        clone: SelfT = self.clone()
        clone._impl_translate(trans_vec)
        return clone
        # return self

    def rotated(self: SelfT, theta: float, origin: Optional[Union[VectorT, str]] = None) -> SelfT:
        """Return a rotated copy around `origin` with angle `theta` in math. positive direction (counterclockwise).

        Args:
            theta (float): rotation angle in rad
            origin (Optional[Union[linalg.VectorLike, str]], optional):
                origin of rotation. If not set, (0,0) is used as
                origin. If origin == 'center', the
                :attr:`Transformable.center` of the object will
                be used. The same applies for the case that
                origin == 'origin' with the
                :attr:`Transformable.origin` property. Default
                to None.

        Returns:
            TransformableBase
        """
        # pylint: disable=protected-access
        if not np.isclose(float(theta), 0.):
            clone: SelfT = self.clone()
            clone._apply_shifted_trafo(clone._impl_rotate, theta, origin)
            return clone
        return self

    def scaled(self: SelfT, fac: float, origin: Optional[Union[VectorT, str]] = None) -> SelfT:
        """Return a scale object homogeneously about `origin` with factor `s`.

        Args:
            fac (float): rotation angle in rad
            origin (Optional[Union[linalg.VectorLike, str]], optional):
                origin of rotation. If not set, (0,0) is used as
                origin. If origin == 'center', the
                :attr:`Transformable.center` of the object will
                be used. The same applies for the case that
                origin == 'origin' with the
                :attr:`Transformable.origin` property. Default
                to None.

        Returns:
            TransformableBase
        """
        if not np.isclose(float(fac), 1.):
            clone: SelfT = self.clone()
            clone._apply_shifted_trafo(clone._impl_scale, fac, origin)  # pylint: disable=protected-access
            return clone
        return self

    def mirrored(self: SelfT, mirror_plane: VectorT) -> SelfT:
        """Return a mirrored object mirrored about `mirror_plane`.

        Args:
            mirror_plane (VectorLike): mirror plane to be used.

        Returns:
            TransformableBase
        """
        clone: T = self.clone()
        clone._impl_mirror(mirror_plane)  # pylint: disable=protected-access
        return clone

    def transformed(self: SelfT, transformations: _TransformationBuilder[VectorT]) -> SelfT:
        """Return a transformed object. the transformation can be build by the following functions:
            - :func:`~fibomat.linalg.transformation_builder.translate`
            - :func:`~fibomat.linalg.transformation_builder.rotate`
            - :func:`~fibomat.linalg.transformation_builder.scale`
            - :func:`~fibomat.linalg.transformation_builder.mirror`

        E.g. ::

            transformable_obj.transform(translate([1, 2]) | rotate(np.pi/3) | mirror([3,4])

        Args:
            transformations (_TransformationBuilder): transformation

        Returns:
            TransformableBase
        """
        # pylint: disable=protected-access
        clone: SelfT = self.clone()
        for trafo in transformations.transformations:
            if isinstance(trafo, _TranslationBuilder):
                clone._impl_translate(trafo.trans_vec)
            elif isinstance(trafo, _RotationBuilder):
                clone._apply_shifted_trafo(clone._impl_rotate, trafo.theta, trafo.origin)
            elif isinstance(trafo, _ScaleBuilder):
                clone._apply_shifted_trafo(clone._impl_scale, trafo.fac, trafo.origin)
            elif isinstance(trafo, _MirrorBuilder):
                clone._impl_mirror(trafo.mirror_plane)
            else:
                raise TypeError(f'{trafo.__class__} is an unknown transforamtion.')

        return clone



"""Provide the :func:`self_intersections` and  :func:`curve_intersections` function."""

from typing import Dict, List, Any

from fibomat.shapes import ArcSpline
from fibomat import _libfibomat


def self_intersections(curve: ArcSpline) -> Dict[str, List[Dict[str, Any]]]:
    """Self intersections of curve.

    Args:
        curve (ArcSpline): curve

    Returns:
        Dict[str, List[Dict[str, Any]]]:
            dict with key 'intersections'. 'intersections' contains list where each element is a dict with keys 'seg_1',
            'seg_2', and 'pos' where the former two elements contain the indices of the segments where the intersection
            occurs and the latter the position of the intersection.
    """
    return {
        'intersections': [
            {'seg_1': s1, 'seg_2': s2, 'pos': pos} for s1, s2, pos in
            _libfibomat.self_intersections(curve.arc_spline_impl)
        ]
    }


def curve_intersections(curve_1: ArcSpline, curve_2: ArcSpline) -> Dict[str, List[Dict[str, Any]]]:
    """Intersections between curves.

    Args:
        curve_1 (ArcSpline): first curve
        curve_2 (ArcSpline): second curve

    .. todo:: what is seg_1, seg_2 in coincidences?

    Returns:
        Dict[str, List[Dict[str, Any]]]:
            dict with key 'intersections' and 'coincidences'. 'intersections' contains list where each element is a dict
            with keys 'seg_1', 'seg_2', and 'pos' where the former two elements contain the indices of the segments
            where the intersection occurs and the latter the position of the intersection. 'coincidences' contains a
            list where each element is a dict with keys 'seg_1', 'seg_2', 'start_pos' and 'end_pos'. 'start_pos' and
            'end_pos' indicate the range where the two curves lie on each other.
    """
    intersection, coincidences = _libfibomat.curve_intersections(curve_1.arc_spline_impl, curve_2.arc_spline_impl)

    return {
        'intersections': [
            {'seg_1': s1, 'seg_2': s2, 'pos': pos} for s1, s2, pos in intersection
        ],
        'coincidences': [
            {'seg_1': s1, 'seg_2': s2, 'start_pos': start, 'end_pos': end} for s1, s2, start, end in coincidences
        ]
    }



"""Provide Curve raster style."""

from fibomat.shapes.rasterizedpoints import RasterizedPoints
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, LengthQuantity
from fibomat.dimensioned_object import DimObjLike, DimObj
from fibomat.mill import Mill
from fibomat.shapes import Shape


class Linear(RasterStyle):
    def __init__(self, pitch_u: LengthQuantity, pitch_v: LengthQuantity, theta: float = 0.):
        """
        Args:
            pitch_u (LengthQuantity): pitch on lines
            pitch_v (LengthQuantity): pitch between lines
            theta (float): rotation angle of lines
        """
        self._pitch_u = pitch_u
        self._pitch_v = pitch_v
        self._theta = theta

    def __repr__(self) -> str:
        return '{}(pitch_u={!r}, pitch_v={!r}, theta={!r})'.format(
            self.__class__.__name__, self._pitch_u, self._pitch_v, self._theta
        )

    @property
    def dimension(self) -> int:
        return 2

    def rasterize(
        self,
        dim_shape: DimObjLike[Shape, LengthUnit],
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPoints:
        raise NotImplementedError



from __future__ import annotations
from typing import TypeVar, overload, Union, Sequence, Iterable, Optional, Any, cast, List
import warnings
import abc

import numpy as np
import pint

FloatTypes = int, float, np.integer, np.float
FloatTypesUnion = Union[int, float, np.integer, np.float]
T = TypeVar('T', int, float, np.integer, np.float, pint.Quantity)  # LengthQuantity
SelfT = TypeVar('SelfT', bound='VectorBase')


class VectorValueError(ValueError):
    """Exception of this type is raised if any non supported value type is passed to Vector.__init__."""


class VectorBase(Sequence[T], abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def _validate_type(val: Any) -> bool:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _build_np_array(val: Sequence[T]) -> np.ndarray:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _make_zero_vector() -> np.ndarray:
        raise NotImplementedError

    @classmethod
    def _from_iterable(cls, x: Iterable[Any]) -> List[T]:  # pylint: disable=invalid-name
        # np.empty(shape=(2,), dtype=float)

        if not hasattr(x, "__iter__"):
            raise VectorValueError("x must be int, float, or iterable")

        i = iter(x)
        try:
            x_val: T = next(i)
            y_val: T = next(i)
        except StopIteration:
            raise VectorValueError('if x is an iterable, it must contain exactly 2 items and not less')

        if not cls._validate_type(x_val):
            raise VectorValueError('x not compatible with vector values.')
        if not cls._validate_type(y_val):
            raise VectorValueError('y not compatible with vector values.')
        try:
            next(i)
            raise VectorValueError('if x is an iterable, it must contain exactly 2 items and not more')
        except StopIteration:
            pass

        return [x_val, y_val]

    @classmethod
    def _from_polar(cls, r: T, phi: float) -> List[T]:  # pylint: disable=invalid-name
        if not cls._validate_type(r):
            raise VectorValueError('r not compatible with vector values.')
        if not isinstance(phi, FloatTypes):
            raise VectorValueError('phi not compatible with vector values.')

        return [r * np.cos(phi), r * np.sin(phi)]

    def __init__(
        self,
        x: Optional[Union[T, VectorBase, Iterable[T]]] = None,
        y: Optional[T] = None,
        r: Optional[T] = None,
        phi: Optional[T] = None,
    ):
        vector: List[T]

        if x is None and y is None and r is None and phi is None:
            vector = self._make_zero_vector()
        elif x is not None:
            if self._validate_type(x):
                if (r is not None) or (phi is not None):
                    raise VectorValueError("Define (x, y) or (r, theta) but not both.")
                if y is not None and self._validate_type(y):
                    x = cast(T, x)
                    vector = self._from_iterable([x, y])
                else:
                    raise VectorValueError('if x is int or float, y must be defined and must have the same type.')
            else:  # x is some kind of list, tuple, Vector, np.ndarray, ...
                if (y is not None) or (r is not None) or (phi is not None):
                    raise VectorValueError("when x is an object, you must not specify y, r, or theta")
                x = cast(Iterable[T], x)
                vector = self._from_iterable(x)
        elif r is not None and phi is not None:  # polar case, r and phi should be defined.
            if y is not None:
                raise VectorValueError("Define (x, y) or (r, theta) but not both.")

            vector = self._from_polar(r, phi)
        else:
            raise VectorValueError('Incompatible combination of arguments.')

        self._vector = self._build_np_array(vector)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(x={self._vector[0]}, y={self._vector[1]})'

    @overload
    def __getitem__(self, index: int) -> T: ...  # pragma: no cover

    @overload
    def __getitem__(self, slice_: slice) -> Sequence[T]: ...  # pragma: no cover

    def __getitem__(self, index: Union[int, slice]) -> Union[T, Sequence[T]]:
        if isinstance(index, int):
            return self._vector[index]
        elif isinstance(index, slice):
            sliced = self._vector[index]
            # needed so that if [T = QuantityType] [Q_(..), Q_(..)] is returned and not Q_([.., ..]).
            return [val for val in sliced]
        else:
            raise TypeError('index must be int or slice.')

    def __len__(self) -> int:
        return 2

    def __array__(self, dtype=None) -> np.ndarray:
        """Helper to allow conversion to numpy array.

        ::

            np_array = np.asarray(Vector(1, 2))

        Returns:
            numpy.ndarray
        """
        # to suppress UnitStrippedWarning in pint module
        with warnings.catch_warnings(record=True):
            return np.array(self._vector, dtype=(dtype or self._vector.dtype))

    @property
    def x(self) -> T:  # pylint: disable=invalid-name
        """X component the vector

        Access:
            get

        Returns:
            T
        """
        return self._vector[0]

    @property
    def y(self) -> T:  # pylint: disable=invalid-name
        """X component the vector

        Access:
            get

        Returns:
            T
        """
        return self._vector[1]

    @property
    def r(self) -> T:  # pylint: disable=invalid-name
        """Radial component the vector.

        Access:
            get

        Returns:
            float
        """
        return self.mag

    @property
    def phi(self) -> float:
        """Angular component of vector (angle between vector and x axis in radiant between -pi and pi).

        Access:
            get

        Returns:
            float
        """
        return float(np.arctan2(self._vector[1], self._vector[0]))

    @property
    def length(self) -> T:
        """Length (magnitude) of vector.

        Access:
            get

        Returns:
            T
        """
        warnings.warn('This method is deprecated. Use mag or magnitude instead', category=DeprecationWarning)
        return np.sqrt((self._vector*self._vector).sum())

    @property
    def magnitude(self) -> T:
        """Magnitude of vector.

        Access:
            get

        Returns:
            T
        """
        return np.sqrt((self._vector*self._vector).sum())

    @property
    def mag(self) -> T:
        """Magnitude of vector (short form of :attr:`VectorBase.magnitude`).

        Access:
            get

        Returns:
            T
        """
        return np.sqrt((self._vector*self._vector).sum())

    @property
    def angle_about_x_axis(self) -> float:
        """Angle between vector and positive x axis (angle will be in [0, 2pi]).

        Access:
            get

        Returns:
            float

        Raises:
            RuntimeError: Raised if self is null vector.
        """
        if np.allclose(self, 0.):
            raise ValueError('Cannot calculate angle of null vector')

        angle = self.phi
        while angle < 0.:
            angle += 2 * np.pi

        assert np.isclose(np.clip(angle, 0., 2 * np.pi), angle)
        return np.clip(angle, 0., 2 * np.pi)

    def close_to(self, other: Iterable[T]) -> bool:
        """Checks if `other` is close to `self` component wise.

        Args:
            other (Vector, Iterable[float]): other vector(like)

        Returns:
            bool
        """
        return np.allclose(self._vector, self.__class__(other))

    def __eq__(self, other: Iterable[T]) -> bool:
        return self.close_to(other)

    def normalized(self: SelfT) -> SelfT:
        """Create a new vector with same :attr:`Vector.phi` but :attr:`Vector.r` = 1.

        Returns:
            SelfT
        """
        return self.__class__(self._vector / np.linalg.norm(self._vector))

    def normalized_to(self: SelfT, length: T) -> SelfT:
        """ Create a new vector :attr:`Vector.phi` but :attr:`Vector.r` = length.

        Args:
            length (float): new length of vector

        Returns:
            Vector
        """
        return self.normalized() * length

    def rotated(self: SelfT, theta: float, origin: Optional[Iterable[T]] = None) -> SelfT:
        """Return a rotated copy the vector around `center` with angle `theta` in math. positive direction.

        Args:
            theta (float): rotation angle in rad
            origin (VectorLike): rotation center, default to [0., 0.]

        Returns:
            Vector
        """
        theta = float(theta)
        cos = np.cos(theta)
        sin = np.sin(theta)

        # m = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

        rotated_vec = self._vector.copy()

        if origin is not None:
            origin_array = self.__class__(origin)._vector

            rotated_vec -= origin_array
            rotated_vec[0] = self._vector[0] * cos - self._vector[1] * sin
            rotated_vec[1] = self._vector[1] * cos + self._vector[0] * sin
            rotated_vec += origin_array
        else:
            rotated_vec[0] = self._vector[0] * cos - self._vector[1] * sin
            rotated_vec[1] = self._vector[1] * cos + self._vector[0] * sin

        return self.__class__(rotated_vec)

    def mirrored(self: SelfT, mirror_axis: Iterable[T]) -> SelfT:
        """Return a mirrored version of the vector.

        Args:
            mirror_axis (VectorLike): mirror axis

        Returns:
            Vector
        """
        # pylint: disable=invalid-name
        mirror_axis = self.__class__(mirror_axis)

        lx, ly = np.asarray(mirror_axis)
        mirror_matrix = np.array([[lx*lx - ly*ly, 2*lx*ly], [2*lx*ly, ly*ly - lx*lx]]) / (lx**2 + ly**2)

        return self.__class__(mirror_matrix @ self._vector)

    def projected(self, other: Iterable[T]) -> SelfT:
        """Project other onto self.
        https://en.wikibooks.org/wiki/Linear_Algebra/Orthogonal_Projection_Onto_a_Line

        Args:
            other:

        Returns:
            Vector
        """
        other = self.__class__(other)

        return self * (self.dot(other) / self.dot(self))

    def dot(self: SelfT, other: Iterable[T]) -> T:
        """Calculate dot product with other vector.

        Args:
            other(VectorLike): other vector

        Returns:
            float
        """
        # pylint: disable=protected-access
        return np.dot(self._vector, self.__class__(other)._vector)

    def cross(self: SelfT, other: Iterable[T]) -> T:
        """Cross product in 2d.

        a x b = a.x * b.y - a.y * b.x

        a = self, b = other

        Args:
            other(VectorLike): other vector

        Returns:
            T
        """
        other = self.__class__(other)

        return self.x * other.y - self.y * other.x

    def __add__(self: SelfT, other: Iterable[T]) -> SelfT:
        """Add operation: self + other.

        Args:
            other (VectorLike): summand

        Returns:
            Vector
        """
        return self.__class__(self._vector + self.__class__(other)._vector)

    def __radd__(self: SelfT, other: Iterable[T]) -> SelfT:
        """Add operation: self + other.

        Args:
            other (VectorLike): summand

        Returns:
            SelfT
        """
        return self.__class__(self._vector + self.__class__(other)._vector)

    def __sub__(self: SelfT, other: Iterable[T]) -> SelfT:
        """Subtraction operation: self - other.

        Args:
            other (VectorLike): subtrahend

        Returns:
            SelfT
        """
        return self.__class__(self._vector - self.__class__(other)._vector)

    def __rsub__(self: SelfT, other: Iterable[T]) -> SelfT:
        """Subtraction operation: other - self.

        Args:
            other (VectorLike): minuend

        Returns:
            SelfT
        """
        return self.__class__(self.__class__(other)._vector - self._vector)

    def __mul__(self: SelfT, other: FloatTypesUnion) -> SelfT:
        """Scalar multiplication: self * other.

        Args:
            other (float): multiplicand

        Returns:
            Vector
        """
        return self.__class__(float(other) * self._vector)

    def __rmul__(self: SelfT, other: FloatTypesUnion) -> SelfT:
        """Scalar multiplication: other * self.

        Args:
            other (float): multiplicand

        Returns:
            SelfT
        """
        return self.__mul__(other)

    def __truediv__(self: SelfT, other: float) -> SelfT:
        """Scalar division: self / other.

        Args:
            other (float): divisor

        Returns:
            SelfT
        """
        return self.__class__(self._vector / float(other))

    def __neg__(self: SelfT) -> SelfT:
        """Negation: -vector.

        Returns:
            SelfT
        """
        return self.__class__(-self._vector)



from typing import Tuple, Sequence

import numpy as np

from fibomat.rasterizedpattern import RasterizedPattern
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, scale_to, has_length_dim, QuantityType, scale_factor
from fibomat.shapes import Shape, DimShape
from fibomat.mill import Mill
from fibomat.curve_tools import rasterize
from fibomat.raster_styles.scansequence import ScanSequence


class Curve(RasterStyle):
    def __init__(self, pitch: QuantityType, scan_sequence: ScanSequence):
        self._pitch = pitch

        if not has_length_dim(self._pitch):
            raise ValueError('Pitch must have [length] as dimension')

        if self._pitch.m <= 0:
            raise ValueError('pitch must be greater than 0.')

        if scan_sequence not in [ScanSequence.CONSECUTIVE, ScanSequence.BACKSTITCH, ScanSequence.BACK_AND_FORTH]:
            raise ValueError(
                'scan_sequence must be "ScanSequence.CONSECUTIVE", " ScanSequence.BACKSTITCH" or '
                '"ScanSequence.BACK_AND_FORTH".'
            )

        self._scan_sequence = scan_sequence

    def __repr__(self):
        return '{}(pitch={!r})'.format(self.__class__.__name__, self._pitch)

    @property
    def dimension(self) -> int:
        return 1

    @property
    def pitch(self) -> QuantityType:
        return self._pitch

    @property
    def scan_sequence(self) -> ScanSequence:
        return self._scan_sequence

    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:
        # dim_shape = DimObj.create(dim_shape)

        pitch_scaled = scale_to(dim_shape.unit, self._pitch)

        points = rasterize(dim_shape.shape, pitch_scaled)

        points._dwell_points[:, :2] *= scale_factor(out_length_unit, dim_shape.unit)
        points._dwell_points[:, 2] = scale_to(out_time_unit, mill.dwell_time)

        if self._scan_sequence in [ScanSequence.BACKSTITCH, ScanSequence.CONSECUTIVE]:
            if self._scan_sequence == ScanSequence.BACKSTITCH:
                # dwell points are indexed by 0, 1, 2, 3, 4, 5, ...
                # with backstiztch, points should be ordered like 1, 0, 3, 2, 5, 4, ...
                # hence, consecutive pairs of points must be swapped
                for i in range(0, points.n_points - 1, 2):
                    points._dwell_points[i], points._dwell_points[i + 1] = (
                        np.array(points._dwell_points[i + 1]), np.array(points._dwell_points[i])
                    )

            return RasterizedPattern(points.repeats_applied(mill.repeats).dwell_points, out_length_unit, out_time_unit)
        elif self._scan_sequence == ScanSequence.BACK_AND_FORTH:
            back_and_force = []
            reversed_points = points.dwell_points[::-1]
            # print(points)
            # print(reversed_points)
            for i in range(mill.repeats):
                if i % 2 == 0:
                    back_and_force.append(points.dwell_points)
                else:
                    back_and_force.append(reversed_points)
            return RasterizedPattern(np.concatenate(back_and_force), out_length_unit, out_time_unit)
        else:
            raise ValueError('Scan sequence not supported.')





"""Provide math helper functions :func:`mod_2pi` and :func:`float_gcd`."""
# pylint: disable=invalid-name
import numpy as np


def mod_2pi(val: float) -> float:
    """Calculate val % 2*pi.

    If val*k % 2*pi == 0, 2*pi is returned for k != 0.

    Args:
        val: value

    Returns:
        float
    """
    mod = np.mod(val, 2*np.pi)
    if np.isclose(mod, 0.):
        if np.isclose(val, 0.):
            return 0.
        return 2*np.pi
    return mod


# https://stackoverflow.com/a/45325587
def float_gcd(a: float, b: float, rtol: float = 1e-05, atol: float = 1e-08):
    """Calculate gcd for float values.

    Args:
        a: first value
        b: second value
        rtol: relative tolerance
        atol: absolut tolerance

    Returns:
        float
    """
    t = min(abs(a), abs(b))
    while abs(b) > rtol * t + atol:
        a, b = b, a % b
    return a
